# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/classes/modelPT.py

import copy
import inspect
import os
import uuid
from abc import abstractmethod
from os import path
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import hydra
import torch
from omegaconf import DictConfig, OmegaConf, open_dict
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.utilities import model_summary, rank_zero_only

import atommic.core.optim
import atommic.utils
from atommic import package_info
from atommic.core.classes.common import Model
from atommic.core.connectors.save_restore_connector import SaveRestoreConnector
from atommic.utils import logging
from atommic.utils.app_state import AppState
from atommic.utils.exceptions import ATOMMICBaseException
from atommic.utils.get_rank import get_rank, is_global_rank_zero

__all__ = ["ModelPT"]


class ModelPT(LightningModule, Model):
    """Interface for Pytorch-lightning based atommic models."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):  # noqa: MC0001
        """Base class from which all atommic models should inherit

        Internal global flags that determine core functionality of ModelPT.

        _MODEL_IS_RESTORED:
            This flag determines the context of the model - whether the model is currently being
            restored or not.

            -   When set, it can be assumed that the model's will disable all automatic methods -
                setup_training_data(), setup_validation/test_data() and their multi equivalents.

            -   If a model is being restored from a archive file (tarfile), it can be assumed that
                under this context, the cwd is *inside* the tarfile itself.

        _MODEL_RESTORE_PATH:
            A string path to a a file from which the model is being restored.

            This file can either be a PyTorch Lightning Checkpoint, or a archive (tarfile) that contains
            artifact objects.

            If it is an archive file, during restoration, the cwd will be temporarily moved to inside the
            archive itself.

        Parameters
        ----------
        cfg : DictConfig
            The cfg object should have (optionally) the following sub-configs:

                - train_ds - to instantiate training dataset

                - validation_ds - to instantiate validation dataset

                - test_ds - to instantiate testing dataset

                - optim - to instantiate optimizer with learning rate scheduler

        trainer : Trainer
            Pytorch Lightning Trainer instance
        """
        if trainer is not None and not isinstance(trainer, Trainer):
            raise ValueError(
                "trainer constructor argument must be either None or pytorch_lightning.Trainer. "
                f"But got {type(trainer)} instead."
            )
        super().__init__()

        # set global vars in AppState
        app_state = AppState()

        # Convert config to a DictConfig
        cfg = atommic.utils.model_utils.convert_model_config_to_dict_config(cfg)

        # Convert config to support Hydra 1.0+ instantiation
        cfg = atommic.utils.model_utils.maybe_update_config_version(cfg)

        if "model" in cfg:
            raise ValueError(
                "Creating model config node is forbidden due to collision problem when loading from checkpoint."
            )

        if "target" not in cfg:
            # This is for Jarvis service.
            OmegaConf.set_struct(cfg, False)
            cfg.target = f"{self.__class__.__module__}.{self.__class__.__name__}"
            OmegaConf.set_struct(cfg, True)

        if "atommic_version" not in cfg:
            with open_dict(cfg):
                cfg.atommic_version = package_info.__version__

        self._cfg = cfg

        # init mapping submodule attribute -> config_field for nested atommic models
        self._atommic_submodule_name_to_config_field: Dict = {}

        self.save_hyperparameters("cfg")
        self._train_dl = None
        self._validation_dl = None
        self._test_dl = None
        self._optimizer_param_groups = None
        self._optimizer = None
        self._scheduler = None
        self.set_trainer(trainer)

        self._save_restore_connector = SaveRestoreConnector()

        self._set_model_guid()

        # Set device_id in AppState
        if torch.cuda.is_available() and torch.cuda.current_device() is not None:
            app_state.device_id = torch.cuda.current_device()

        if self._cfg is not None and not self._is_model_being_restored():
            # Setup data loaders now (default) or defer setup to `self.setup()` if `defer_setup` is set in the config
            # of the corresponding dataloader.
            if (
                "train_ds" in self._cfg
                and self._cfg.train_ds is not None
                and not self._cfg.train_ds.get("defer_setup", False)
            ):
                self.setup_training_data(self._cfg.train_ds)

            if (
                "validation_ds" in self._cfg
                and self._cfg.validation_ds is not None
                and not self._cfg.validation_ds.get("defer_setup", False)
            ):
                self.setup_multiple_validation_data(val_data_config=cfg.validation_ds)

            if (
                "test_ds" in self._cfg
                and self._cfg.test_ds is not None
                and not self._cfg.test_ds.get("defer_setup", False)
            ):
                self.setup_multiple_test_data(test_data_config=cfg.test_ds)

        else:
            if "train_ds" in self._cfg and self._cfg.train_ds is not None:
                logging.warning(
                    "If you intend to do training or fine-tuning, please call the ModelPT.setup_training_data() "
                    "method and provide a valid configuration file to setup the train data loader.\n"
                    f"Train config : \n{OmegaConf.to_yaml(self._cfg.train_ds)}"
                )
            if "validation_ds" in self._cfg and self._cfg.validation_ds is not None:
                logging.warning(
                    "If you intend to do validation, please call the ModelPT.setup_validation_data() or "
                    "ModelPT.setup_multiple_validation_data() method and provide a valid configuration file to "
                    "setup the validation data loader(s). \n"
                    f"Validation config : \n{OmegaConf.to_yaml(self._cfg.validation_ds)}"
                )
            if "test_ds" in self._cfg and self._cfg.test_ds is not None:
                logging.warning(
                    "Please call the ModelPT.setup_test_data() or ModelPT.setup_multiple_test_data() method "
                    "and provide a valid configuration file to setup the test data loader(s).\n"
                    f"Test config : \n{OmegaConf.to_yaml(self._cfg.test_ds)}"
                )

        # Create list of lists for val and test outputs to support multiple dataloaders
        # Initialize an empty list as sometimes self._validation_dl can be None at this stage
        self.validation_step_outputs: List = []
        # Check len(self._validation_dl) > 1 as sometimes single dataloader can be in a list: [<Dataloader obj>]
        # when ds_item in config has 1 item passed in a list
        if self._validation_dl and isinstance(self._validation_dl, list) and len(self._validation_dl) > 1:
            for _ in range(len(self._validation_dl)):
                self.validation_step_outputs.append([])

        # Initialize an empty list as sometimes self._test_dl can be None at this stage
        self.test_step_outputs: List = []
        if self._test_dl and isinstance(self._test_dl, list) and len(self._test_dl) > 1:
            for _ in range(len(self._test_dl)):
                self.test_step_outputs.append([])
        # ModelPT wrappers over subclass implementations
        self.training_step = atommic.utils.model_utils.wrap_training_step(  # pylint: disable=no-value-for-parameter
            self.training_step  # type: ignore
        )

        # Setup nsys profiling if it has been enabled in the model config
        self._setup_nsys_profiling()

    def __init_subclass__(cls) -> None:
        """This method is called when a subclass is created."""
        cls._save_restore_connector = SaveRestoreConnector()

    def register_artifact(self, config_path: str, src: str, verify_src_exists: bool = True):
        r"""Register model artifacts with this function. These artifacts (files) will be included inside .atommic file
        when model.save_to("model.atommic") is called.

        How it works:

            1. It always returns existing absolute path which can be used during Model constructor call EXCEPTION: src
                is None or "" in which case nothing will be done and src will be returned

            2. It will add (config_path, model_utils.ArtifactItem()) pair to self.artifacts

            .. code-block::

                    If "src" is local existing path:

                        then it will be returned in absolute path form.

                    elif "src" starts with "atommic_file:unique_artifact_name":

                        .atommic will be untarred to a temporary folder location and an actual existing path will be
                        returned

                    else:

                        an error will be raised.

        If "src" is local existing path, then it will be returned in absolute path form.
        elif "src" starts with "atommic_file:unique_artifact_name" .atommic will be untarred to a temporary folder
        location and an actual existing path will be returned else an error will be raised.

        WARNING: use .register_artifact calls in your models' constructors.
        The returned path is not guaranteed to exist after you have exited your model's constructor.

        Parameters
        ----------
        config_path : str
            Artifact key. Usually corresponds to the model config.
        src : str
            Path to artifact.
        verify_src_exists : bool
            If set to False, then the artifact is optional and register_artifact will return None even if src is not
            found. Default is ``True``.

        Returns
        -------
        If src is not None or empty it always returns absolute path which is guaranteed to exist during model instance
        life.
        """
        if src is None or not src:
            return src

        if Path(src).suffix == ".atommic":
            raise ATOMMICBaseException(
                "Registering .atommic files as artifacts not supported. "
                "If you are trying to make a nested model, use `register_atommic_submodule`."
            )

        if not hasattr(self, "artifacts"):
            self.artifacts: Dict[str, atommic.utils.model_utils.ArtifactItem] = {}

        if self.artifacts is None:
            self.artifacts = {}

        if config_path in self.artifacts:
            logging.warning(
                f"You tried to register an artifact under config key={config_path} but an artifact for "
                "it has already been registered."
            )

        return self._save_restore_connector.register_artifact(self, config_path, src, verify_src_exists)

    def has_artifacts(self) -> bool:
        """Returns True if model has artifacts registered."""
        return hasattr(self, "artifacts") and self.artifacts is not None and len(self.artifacts) > 0

    def has_native_or_submodules_artifacts(self) -> bool:
        """Returns True if it has artifacts or any of the submodules have artifacts."""
        return any(
            (
                isinstance(module, ModelPT)
                and hasattr(module, "artifacts")
                and module.artifacts is not None
                and len(module.artifacts) > 0
            )
            for module in self.modules()
        )

    def has_atommic_submodules(self) -> bool:
        """Returns True if it has any registered atommic submodules."""
        return len(self._atommic_submodule_name_to_config_field) > 0

    def register_atommic_submodule(self, name: str, config_field: str, model: "ModelPT") -> None:
        """Adds a atommic model as a submodule.

        Submodule can be accessed via the `name` attribute on the parent atommic model this submodule was registered on
        (`self`).

        In the saving process, the whole parent model (self) is held as a solid model with artifacts from the child
        submodule, the submodule config will be saved to the `config_field` of the parent model.

        This method is necessary to create a nested model, e.g.
            .. code-block:: python

                class ParentModel(ModelPT):
                    def __init__(self, cfg, trainer=None):
                        super().__init__(cfg=cfg, trainer=trainer)

                        # annotate type for autocompletion and type checking (optional)
                        self.child_model: Optional[ChildModel] = None
                        if cfg.get("child_model") is not None:
                            self.register_atommic_submodule(
                                name="child_model",
                                config_field="child_model",
                                model=ChildModel(self.cfg.child_model, trainer=trainer),
                            )
                        # ... other code

        Parameters
        ----------
        name : str
            Name of the submodule. This name will be used to access the submodule from the parent model.
        config_field : str
            Name of the config field where the submodule config will be saved.
        model : ModelPT
            The submodule model.
        """
        # check it is a real atommic model
        if not isinstance(model, ModelPT):
            raise ATOMMICBaseException(
                f"Model is not and instance of ModelPT, so can't be registered. Got {type(model).__name__}"
            )
        # check if it is called after __init__
        if not hasattr(self, "_atommic_submodule_name_to_config_field"):
            raise ATOMMICBaseException(
                "You are trying to register a submodule before the model is initialized. This is not allowed. "
                "Did you forget to call `super().__init__`?"
            )
        # assign attribute to self
        setattr(self, name, model)
        # add to the submodules mapping
        self._atommic_submodule_name_to_config_field[name] = config_field

    def named_atommic_modules(
        self, prefix_name: str = "", prefix_config: str = ""
    ) -> Iterator[Tuple[str, str, "ModelPT"]]:
        """Returns an iterator over all atommic submodules recursively, yielding tuples of (attribute path, path in
        config, submodule), starting from the core module.

        Parameters
        ----------
        prefix_name : str
            Prefix for the name path.
        prefix_config : str
            Prefix for the path in config.

        Returns
        -------
        Iterator over (attribute path, path in config, submodule), starting from (prefix, self).
        """
        if not hasattr(self, "_atommic_submodule_name_to_config_field"):
            raise ATOMMICBaseException(
                "Model is not fully initialized. Calling `named_atommic_modules` before __init__ not allowed. "
                "Did you forget to call `super().__init__`?"
            )

        yield prefix_name, prefix_config, self

        # recursive iteration over all atommic submodules
        for name, config_field in self._atommic_submodule_name_to_config_field.items():
            attribute_path = f"{prefix_name}.{name}" if prefix_name else name
            config_path = f"{prefix_config}.{config_field}" if prefix_config else config_field
            module: ModelPT = getattr(self, name)
            for submodule_name, subconfig_path, submodule in module.named_atommic_modules(
                prefix_name=attribute_path, prefix_config=config_path
            ):
                yield submodule_name, subconfig_path, submodule

    def save_to(self, save_path: str):
        """Saves model instance (weights and configuration) into .atommic file. You can use "restore_from" method to
        fully restore instance from .atommic file. .atommic file is an archive (tar.gz) with the following:

            * model_config.yaml - model configuration in .yaml format. You can deserialize this into cfg argument for
            model's constructor

            * model_wights.ckpt - model checkpoint

        Parameters
        ----------
        Path to .atommic file where model instance should be saved.
        """

        def maybe_make_save_dir(_path: "Path"):
            """Creates directory if it does not exist"""
            if not _path.parent.exists():
                _path.parent.mkdir(parents=True)

        save_path = Path(save_path).expanduser().resolve()  # type: ignore
        app_state = AppState()
        if app_state.model_parallel_size is not None:
            if app_state.model_parallel_size > 1 and isinstance(self._save_restore_connector, SaveRestoreConnector):
                raise ValueError(
                    "Default atommic SaveRestoreConnector will not work in model parallel mode. You should use a "
                    "connector which supports model parallel mode. You can also use a custom one."
                )
            if is_global_rank_zero():
                maybe_make_save_dir(save_path)  # type: ignore
            if torch.distributed.is_initialized():
                torch.distributed.barrier()
            # connector checks for ranks properly, no need to check here
            # downstream tasks expect str, not Path
            self._save_restore_connector.save_to(self, str(save_path))
        elif is_global_rank_zero():
            maybe_make_save_dir(Path(save_path))
            # downstream tasks expect str, not Path
            self._save_restore_connector.save_to(self, str(save_path))

    @classmethod
    def restore_from(  # type: ignore  # pylint: disable=arguments-renamed
        cls,
        restore_path: str,
        override_config_path: Optional[Union[OmegaConf, str]] = None,
        map_location: Optional[torch.device] = None,
        strict: bool = True,
        return_config: bool = False,
        save_restore_connector: SaveRestoreConnector = None,
        trainer: Optional[Trainer] = None,
    ):
        """Restores model instance (weights and configuration) from .atommic file.

        Parameters
        ----------
        restore_path : str
            Path to .atommic file from which model should be instantiated override_config_path: path to a yaml config
            that will override the internal config file or an OmegaConf/DictConfig object representing the model
            config.
        override_config_path : Optional[Union[OmegaConf, str]]
            path to a yaml config that will override the internal config file or an OmegaConf/DictConfig object
            representing the model config.
        map_location : Optional[torch.device]
            Optional torch.device() to map the instantiated model to a device. Default is ``None``, it will select a
            GPU if available, falling back to CPU otherwise.
        strict : bool
            Passed to load_state_dict. Default is ``True``.
        return_config : bool
            If set to true, will return just the underlying config of the restored model as an OmegaConf/DictConfig
            object without instantiating the model.
        save_restore_connector : SaveRestoreConnector
            Can be overridden to add custom save and restore logic.
        trainer : Optional[Trainer]
            Optional, a pytorch lightning Trainer object that will be forwarded to the instantiated model's
            constructor.

        Example
        -------

        .. code-block::

            model = atommic.collections.reconstruction.nn.CIRIM.restore_from('CIRIM.atommic')
            assert isinstance(model, atommic.collections.reconstruction.nn.CIRIM)

        Returns
        -------
        An instance of type cls or its underlying config (if return_config is set).
        """
        if save_restore_connector is None:
            save_restore_connector = SaveRestoreConnector()

        if save_restore_connector.model_extracted_dir is None:
            restore_path = os.path.abspath(os.path.expanduser(restore_path))
        else:
            restore_path = os.path.abspath(os.path.expanduser(save_restore_connector.model_extracted_dir))

        if not path.exists(restore_path):
            raise FileNotFoundError(f"Can't find {restore_path}")

        app_state = AppState()
        app_state.model_restore_path = restore_path

        cls.update_save_restore_connector(save_restore_connector)
        instance = cls._save_restore_connector.restore_from(
            cls, restore_path, override_config_path, map_location, strict, return_config, trainer
        )
        if isinstance(instance, ModelPT):
            instance._save_restore_connector = save_restore_connector  # pylint: disable=protected-access
        return instance

    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path: str,
        *args,
        map_location: Optional[Union[Dict[str, str], str, torch.device, int, Callable]] = None,
        hparams_file: Optional[str] = None,
        strict: bool = True,
        **kwargs,
    ):
        """Loads ModelPT from checkpoint, with some maintenance of restoration. For documentation, please refer to
        LightningModule.load_from_checkpoint() documentation.

        Parameters
        ----------
        checkpoint_path : str
            Path to checkpoint. Will be passed to LightningModule.load_from_checkpoint().
        *args
            Will be passed to LightningModule.load_from_checkpoint().
        map_location : Optional[Union[Dict[str, str], str, torch.device, int, Callable]]
            Will be passed to LightningModule.load_from_checkpoint().
        hparams_file : Optional[str]
            Will be passed to LightningModule.load_from_checkpoint().
        strict : bool
            Will be passed to LightningModule.load_from_checkpoint().
        """
        checkpoint = None

        try:
            cls._set_model_restore_state(is_being_restored=True)

            checkpoint = super().load_from_checkpoint(
                checkpoint_path=checkpoint_path,
                *args,
                map_location=map_location,
                hparams_file=hparams_file,
                strict=strict,
                **kwargs,
            )

        finally:
            cls._set_model_restore_state(is_being_restored=False)
        return checkpoint

    @abstractmethod
    def setup_training_data(self, train_data_config: Union[DictConfig, Dict]):
        """Setups data loader to be used in training."""

    @abstractmethod
    def setup_validation_data(self, val_data_config: Union[DictConfig, Dict]):
        """Setups data loader to be used in validation."""

    def setup_test_data(self, test_data_config: Union[DictConfig, Dict]):
        """(Optionally) Setups data loader to be used in test."""
        raise NotImplementedError()

    def setup_multiple_validation_data(self, val_data_config: Union[DictConfig, Dict]):
        """(Optionally) Setups data loader to be used in validation."""
        # Set some placeholder overridden by helper method
        self._val_dl_idx = 0
        self.validation_names = None

        # preserve config
        self._update_dataset_config(dataset_name="validation", config=val_data_config)

        try:
            self._multi_dataset_mode = True
            atommic.utils.model_utils.resolve_validation_dataloaders(model=self)
        finally:
            self._multi_dataset_mode = False

        if (
            self.validation_names is None
            and self._validation_dl is not None
            and type(self._validation_dl) in [list, tuple]
        ):
            self.validation_names = [f"val_{idx}_" for idx in range(len(self._validation_dl))]

    def setup_multiple_test_data(self, test_data_config: Union[DictConfig, Dict]):
        """(Optionally) Setups data loader to be used in test, with support for multiple data loaders."""
        # Set some placeholder overridden by helper method
        self._test_dl_idx = 0
        self.test_names = None
        self._test_dl = None

        # preserve config
        self._update_dataset_config(dataset_name="test", config=test_data_config)

        try:
            self._multi_dataset_mode = True
            atommic.utils.model_utils.resolve_test_dataloaders(model=self)
        finally:
            self._multi_dataset_mode = False

        if self.test_names is None and self._test_dl is not None and type(self._test_dl) in [list, tuple]:
            self.test_names = [f"test_{idx}_" for idx in range(len(self._test_dl))]

    def setup_optimization(self, optim_config: Optional[Union[DictConfig, Dict]] = None):  # noqa: MC0001
        """Prepares an optimizer from a string name and its optional config parameters.

        Parameters
        ----------
        optim_config : Optional[Union[DictConfig, Dict]]
            A dictionary containing the following keys:
                - lr : float
                    Mandatory key for learning rate. Will raise ValueError if not provided.
                - optimizer : str
                    String name pointing to one of the available optimizers in the registry. Default is ``Adam``.
                - opt_args : list
                    Optional list of strings, in the format "arg_name=arg_value". The list of "arg_value" will be
                    parsed and a dictionary of optimizer kwargs will be built and supplied to instantiate the
                    optimizer.

        Returns
        -------
        An instance of an optimizer.
        """
        # Setup the optimizer parameter groups (by default use all parameters that are trainable).
        self.setup_optimizer_param_groups()

        # If config was not explicitly provided, use default
        if optim_config is None and self._cfg is not None and hasattr(self._cfg, "optim"):
            optim_config = self._cfg.optim

        # If config is still None, or internal config has no Optim, return without instantiation
        if optim_config is None:
            logging.info("No optimizer config provided, therefore no optimizer was created")
            return
        # Preserve the configuration
        if not isinstance(optim_config, DictConfig):
            optim_config = OmegaConf.create(optim_config)

        # See if internal config has `optim` namespace before preservation
        if self._cfg is not None and hasattr(self._cfg, "optim"):
            if self._cfg.optim is None:
                self._cfg.optim = copy.deepcopy(optim_config)
            else:
                with open_dict(self._cfg.optim):
                    self._cfg.optim = copy.deepcopy(optim_config)

        # Setup optimizer and scheduler
        if optim_config is not None and isinstance(optim_config, DictConfig):
            optim_config = OmegaConf.to_container(optim_config, resolve=True)

        if self._trainer is None:
            logging.warning("Trainer wasn't specified in model constructor. Make sure that you really wanted it.")

        if "sched" in optim_config and self._trainer is not None:
            if not isinstance(self._trainer.accumulate_grad_batches, int):
                raise ValueError("We do not currently support gradient accumulation that is not an integer.")
            if self._trainer.max_steps is None or self.trainer.max_steps < 0:
                # Store information needed to calculate max_steps
                optim_config["sched"]["t_max_epochs"] = self._trainer.max_epochs
                optim_config["sched"]["t_accumulate_grad_batches"] = self._trainer.accumulate_grad_batches
                optim_config["sched"]["t_limit_train_batches"] = self._trainer.limit_train_batches
                optim_config["sched"]["t_num_workers"] = self._trainer.num_devices * self._trainer.num_nodes
            else:
                optim_config["sched"]["max_steps"] = self._trainer.max_steps

        # Force into DictConfig from nested structure
        optim_config = OmegaConf.create(optim_config)
        # Get back nested dict so we its mutable
        optim_config = OmegaConf.to_container(optim_config, resolve=True)

        # Extract scheduler config if inside optimizer config
        if "sched" in optim_config:
            scheduler_config = optim_config.pop("sched")
        else:
            scheduler_config = None

        # Check if caller provided optimizer name, default to Adam otherwise
        optimizer_cls = optim_config.get("_target_", None)

        if optimizer_cls is None:
            # Try to get optimizer name for dynamic resolution, defaulting to Adam
            optimizer_name = optim_config.get("name", "adam")
        elif inspect.isclass(optimizer_cls):
            optimizer_name = optimizer_cls.__name__.lower()
        else:
            # resolve the class name (lowercase) from the class path if not provided
            optimizer_name = optimizer_cls.split(".")[-1].lower()

        # We are guaranteed to have lr since it is required by the argparser
        # But maybe user forgot to pass it to this function
        lr = optim_config.get("lr", None)

        # Check if caller has optimizer kwargs, default to empty dictionary
        if "args" in optim_config:
            optimizer_args = optim_config.pop("args")
            optimizer_args = atommic.core.optim.optimizers.parse_optimizer_args(optimizer_name, optimizer_args)
        else:
            optimizer_args = copy.deepcopy(optim_config)

            # Remove extra parameters from optimizer_args nest
            # Assume all other parameters are to be passed into optimizer constructor
            optimizer_args.pop("name", None)
            optimizer_args.pop("cls", None)
            optimizer_args.pop("lr", None)

        # Adaptive schedulers don't need `lr`
        if lr is not None:
            optimizer_args["lr"] = lr

            # Actually instantiate the optimizer
            if optimizer_cls is None:
                optimizer = atommic.core.optim.optimizers.get_optimizer(optimizer_name)
                optimizer = optimizer(self._optimizer_param_groups, **optimizer_args)

                logging.info("Optimizer config = %s", str(optimizer))

                self._optimizer = optimizer  # type: ignore

            elif inspect.isclass(optimizer_cls):
                optimizer = optimizer_cls(self._optimizer_param_groups, **optimizer_args)
                logging.info("Optimizer config = %s", str(optimizer))

                self._optimizer = optimizer  # type: ignore

            else:
                # Attempt class path resolution
                try:
                    optimizer_cls = OmegaConf.create({"_target_": optimizer_cls})
                    optimizer_config = {"lr": lr} if lr is not None else {}
                    optimizer_config |= optimizer_args

                    optimizer_instance = hydra.utils.instantiate(
                        optimizer_cls, self._optimizer_param_groups, **optimizer_config
                    )  # type: DictConfig

                    logging.info("Optimizer config = %s", str(optimizer_instance))

                    self._optimizer = optimizer_instance

                except Exception as e:
                    logging.error(
                        f"Could not instantiate class path - {optimizer_cls} with kwargs {str(optimizer_config)}"
                    )

                    raise e

            # print(f"scheduler_config = {scheduler_config}")
            if isinstance(scheduler_config["name"], list):
                _schedulers = [
                    atommic.core.optim.lr_scheduler.prepare_lr_scheduler(
                        optimizer=self._optimizer,
                        scheduler_config={
                            "name": scheduler_config["name"][i],
                            "min_lr": scheduler_config["min_lr"][i],
                            "last_epoch": scheduler_config["last_epoch"][i],
                            "warmup_ratio": scheduler_config["warmup_ratio"][i],
                            "monitor": scheduler_config["monitor"][i],
                            "t_max_epochs": scheduler_config["t_max_epochs"],
                            "t_accumulate_grad_batches": scheduler_config["t_accumulate_grad_batches"],
                            "t_limit_train_batches": scheduler_config["t_limit_train_batches"],
                            "t_num_workers": scheduler_config["t_num_workers"],
                        },
                        train_dataloader=self._train_dl,
                    )
                    for i in range(len(scheduler_config["name"]))
                ]

                self._scheduler = _schedulers  # type: ignore
                self._optimizer = [self._optimizer] * len(scheduler_config["name"])  # type: ignore
            else:
                # Try to instantiate scheduler for optimizer
                self._scheduler = atommic.core.optim.lr_scheduler.prepare_lr_scheduler(  # type: ignore
                    optimizer=self._optimizer, scheduler_config=scheduler_config, train_dataloader=self._train_dl
                )

            # Return the optimizer with/without scheduler. This return allows multiple optimizers or schedulers to be
            # created
            return self._optimizer, self._scheduler

    def setup_optimizer_param_groups(self):
        """Used to create param groups for the optimizer. As an example, this can be used to specify per-layer learning
        rates:

        .. code-block::

            optim.SGD([
                        {'params': model.base.parameters()},
                        {'params': model.classifier.parameters(), 'lr': 1e-3}
                        ], lr=1e-2, momentum=0.9)

        See https://pytorch.org/docs/stable/optim.html for more information. By default, ModelPT will use
        self.parameters(). Override this method to add custom param groups.
        """
        param_groups = None
        if hasattr(self, "parameters"):
            param_groups = [{"params": self.parameters()}]
        self._optimizer_param_groups = param_groups

    def configure_optimizers(self):
        """Configure optimizers and schedulers for training."""
        self.setup_optimization()

        if isinstance(self._scheduler, list) and self._scheduler[0] is None:
            return self._optimizer

        if self._scheduler is None:
            return self._optimizer

        if isinstance(self._optimizer, list):
            return self._optimizer, self._scheduler

        return [self._optimizer], [self._scheduler]

    def setup(self, stage: Optional[str] = None):
        """Called at the beginning of fit, validate, test, or predict. This is called on every process when using DDP.

        Parameters
        ----------
        stage : str
            fit, validate, test or predict
        """
        if stage == "fit":
            train_deferred_setup = (
                "train_ds" in self._cfg
                and self._cfg.train_ds is not None
                and self._cfg.train_ds.get("defer_setup", False)
            )
            if self.train_dataloader() is None and train_deferred_setup:
                self.setup_training_data(self._cfg.train_ds)

        if stage in ("fit", "validate"):
            val_deferred_setup = (
                "validation_ds" in self._cfg
                and self._cfg.validation_ds is not None
                and self._cfg.validation_ds.get("defer_setup", False)
            )
            if self.val_dataloader() is None and val_deferred_setup:
                self.setup_multiple_validation_data(val_data_config=self._cfg.validation_ds)

        if stage == "test":
            test_deferred_setup = (
                "test_ds" in self._cfg
                and self._cfg.test_ds is not None
                and self._cfg.test_ds.get("defer_setup", False)
            )
            if self.test_dataloader() is None and test_deferred_setup:
                self.setup_multiple_test_data(test_data_config=self._cfg.test_ds)

    def train_dataloader(self):
        """Return the training dataloader."""
        return self._train_dl if self._train_dl is not None else None

    def val_dataloader(self):
        """Return the validation dataloader."""
        if self._validation_dl is None:
            # None dataloader no longer supported in PTL2.0
            self._validation_dl = []
        return self._validation_dl

    def test_dataloader(self):
        """Return the test dataloader."""
        if self._test_dl is None:
            # None dataloader no longer supported in PTL2.0
            self._test_dl = []
        return self._test_dl

    def on_validation_epoch_end(self) -> Optional[Dict[str, Dict[str, torch.Tensor]]]:
        """Default DataLoader for Validation set which automatically supports multiple data loaders via
        `multi_validation_epoch_end`. If multi dataset support is not required, override this method entirely in base
        class. In such a case, there is no need to implement `multi_validation_epoch_end` either.

        .. note::
            If more than one data loader exists, and they all provide `val_loss`,
            only the `val_loss` of the first data loader will be used by default.
            This default can be changed by passing the special key `val_dl_idx: int`
            inside the `validation_ds` config.

        Returns
        -------
        A dictionary containing the union of all items from individual data_loaders, along with merged logs from all
        data loaders.
        """
        # Case where we dont provide data loaders
        if self.validation_step_outputs is not None and len(self.validation_step_outputs) == 0:
            return {}

        # Case where we provide exactly 1 data loader
        if isinstance(self.validation_step_outputs[0], dict):
            output_dict = self.multi_validation_epoch_end(self.validation_step_outputs, dataloader_idx=0)

            if output_dict is not None and 'log' in output_dict:
                self.log_dict(output_dict.pop('log'), on_epoch=True)

            self.validation_step_outputs.clear()  # free memory
            return output_dict

        # Case where we provide more than 1 data loader
        output_dict = {'log': {}}

        # The output is a list of list of dicts, outer list corresponds to dataloader idx
        for dataloader_idx, val_outputs in enumerate(self.validation_step_outputs):
            # Get prefix and dispatch call to multi epoch end
            dataloader_prefix = self.get_validation_dataloader_prefix(dataloader_idx)
            dataloader_logs = self.multi_validation_epoch_end(val_outputs, dataloader_idx=dataloader_idx)

            # If result was not provided, generate empty dict
            dataloader_logs: Dict[Any, Any] = dataloader_logs or {}  # type: ignore

            # Perform `val_loss` resolution first (if provided outside logs)
            if ("val_loss" in dataloader_logs and "val_loss" not in output_dict) and (  # type: ignore
                dataloader_idx == self._val_dl_idx
            ):
                output_dict["val_loss"] = dataloader_logs["val_loss"]  # type: ignore

            # For every item in the result dictionary
            for k, v in dataloader_logs.items():  # type: ignore
                # If the key is `log`
                if k == "log":
                    # Parse every element of the log, and attach the prefix name of the data loader
                    log_dict = {}

                    for k_log, v_log in v.items():
                        # If we are logging the metric, but dont provide it at result level, store it twice - once in
                        # log and once in result level. Also mark log with prefix name to avoid log level clash with
                        # other data loaders.
                        if k_log not in output_dict["log"] and dataloader_idx == self._val_dl_idx:  # type: ignore
                            new_k_log = k_log
                            # Also insert duplicate key with prefix for ease of comparison / avoid name clash
                            log_dict[dataloader_prefix + k_log] = v_log
                        else:
                            # Simply prepend prefix to key and save
                            new_k_log = dataloader_prefix + k_log

                        # Store log value
                        log_dict[new_k_log] = v_log

                    # Update log storage of individual data loader
                    output_logs = output_dict["log"]  # type: ignore
                    output_logs.update(log_dict)

                    # Update global log storage
                    output_dict["log"] = output_logs  # type: ignore

                else:
                    # If any values are stored outside 'log', simply prefix name and store
                    new_k = dataloader_prefix + k
                    output_dict[new_k] = v  # type: ignore

            self.validation_step_outputs[dataloader_idx].clear()  # free memory

        if "log" in output_dict:  # type: ignore
            self.log_dict(output_dict.pop("log"), on_epoch=True)  # type: ignore

        # return everything else
        return output_dict

    def on_test_epoch_end(self) -> Optional[Dict[str, Dict[str, torch.Tensor]]]:
        """Default DataLoader for Test set which automatically supports multiple data loaders via
        `multi_test_epoch_end`. If multi dataset support is not required, override this method entirely in base class.
        In such a case, there is no need to implement `multi_test_epoch_end` either.

        .. note::
            If more than one data loader exists, and they all provide `test_loss`,
            only the `test_loss` of the first data loader will be used by default.
            This default can be changed by passing the special key `_test_dl_idx: int`
            inside the `test_ds` config.

        Returns
        -------
        Dict[str, Dict[str, torch.Tensor]]
            A dictionary containing the union of all items from individual data_loaders, along with merged logs from
            all data loaders.
        """
        # Case where we dont provide data loaders
        if self.test_step_outputs is not None and len(self.test_step_outputs) == 0:
            return {}

        # Case where we provide exactly 1 data loader
        if isinstance(self.test_step_outputs[0], dict):
            output_dict = self.multi_test_epoch_end(self.test_step_outputs, dataloader_idx=0)

            if output_dict is not None and 'log' in output_dict:
                self.log_dict(output_dict.pop('log'), on_epoch=True)

            self.test_step_outputs.clear()  # free memory
            return output_dict

        # Case where we provide more than 1 data loader
        output_dict = {'log': {}}

        # The output is a list of list of dicts, outer list corresponds to dataloader idx
        for dataloader_idx, test_outputs in enumerate(self.test_step_outputs):
            # Get prefix and dispatch call to multi epoch end
            dataloader_prefix = self.get_test_dataloader_prefix(dataloader_idx)
            self.multi_test_epoch_end(test_outputs, dataloader_idx=dataloader_idx)

            # If result was not provided, generate empty dict
            dataloader_logs = dataloader_logs or {}  # type: ignore  # noqa: F821

            # Perform `test_loss` resolution first (if provided outside logs)
            if (
                "test_loss" in dataloader_logs
                and "test_loss" not in output_dict  # type: ignore
                and dataloader_idx == self._test_dl_idx
            ):
                output_dict["test_loss"] = dataloader_logs["test_loss"]  # type: ignore

            # For every item in the result dictionary
            for k, v in dataloader_logs.items():
                # If the key is `log`
                if k == "log":
                    # Parse every element of the log, and attach the prefix name of the data loader
                    log_dict = {}
                    for k_log, v_log in v.items():
                        # If we are logging the loss, but dont provide it at result level,
                        # store it twice - once in log and once in result level.
                        # Also mark log with prefix name to avoid log level clash with other data loaders
                        if k_log not in output_dict["log"] and dataloader_idx == self._test_dl_idx:  # type: ignore
                            new_k_log = k_log
                            # Also insert duplicate key with prefix for ease of comparison / avoid name clash
                            log_dict[dataloader_prefix + k_log] = v_log
                        else:
                            # Simply prepend prefix to key and save
                            new_k_log = dataloader_prefix + k_log
                        log_dict[new_k_log] = v_log

                    # Update log storage of individual data loader
                    output_logs = output_dict.get("log", {})  # type: ignore
                    output_logs.update(log_dict)

                    # Update global log storage
                    output_dict["log"] = output_logs  # type: ignore

                else:
                    # If any values are stored outside 'log', simply prefix name and store
                    new_k = dataloader_prefix + k
                    output_dict[new_k] = v  # type: ignore

            self.test_step_outputs[dataloader_idx].clear()  # free memory

        if "log" in output_dict:  # type: ignore
            self.log_dict(output_dict.pop("log"), on_epoch=True)  # type: ignore

        # return everything else
        return output_dict

    @staticmethod
    def multi_validation_epoch_end(
        outputs: Union[object, List[Dict[str, torch.Tensor]], None],  # pylint: disable=unused-argument
        dataloader_idx: int = 0,  # pylint: disable=unused-argument
    ) -> None:
        """Adds support for multiple validation datasets. Should be overridden by subclass, to obtain appropriate logs
        for each of the dataloaders.

        Parameters
        ----------
        outputs : Union[List[Dict[str, torch.Tensor]], List[List[Dict[str, torch.Tensor]]]]
            Same as that provided by LightningModule.on_validation_epoch_end() for a single dataloader.
        dataloader_idx : int
            The index of the dataloader for which the outputs are provided.

        Returns
        -------
        A dictionary of values, optionally containing a sub-dict `log`, such that the values in the log will be
        pre-pended by the dataloader prefix.
        """
        logging.warning(
            "Multi data loader support has been enabled, but `multi_validation_epoch_end(outputs, dataloader_idx) "
            "has not been implemented.\n"
            "If you require multi data loader support for validation sets, please override this method.\n"
            "If you do not require multi data loader support, please instead override "
            "`on_validation_epoch_end(outputs)."
        )

    @staticmethod
    def multi_test_epoch_end(
        outputs: Union[object, List[Dict[str, torch.Tensor]]],  # pylint: disable=unused-argument
        dataloader_idx: int = 0,  # pylint: disable=unused-argument
    ) -> None:
        """Adds support for multiple test datasets. Should be overridden by subclass, to obtain appropriate logs for
        each of the dataloaders.

        Parameters
        ----------
        outputs : Union[List[Dict[str, torch.Tensor]], List[List[Dict[str, torch.Tensor]]]]
            Same as that provided by LightningModule.on_test_epoch_end() for a single dataloader.
        dataloader_idx : int
            The index of the dataloader for which the outputs are provided.

        Returns
        -------
        A dictionary of values, optionally containing a sub-dict `log`, such that the values in the log will be
        pre-pended by the dataloader prefix.
        """
        logging.warning(
            "Multi data loader support has been enabled, but `multi_test_epoch_end(outputs, dataloader_idx) has not "
            "been implemented.\n"
            "If you require multi data loader support for validation sets, please override this method.\n"
            "If you do not require multi data loader support, please instead override on_test_epoch_end(outputs)."
        )

    def get_validation_dataloader_prefix(self, dataloader_idx: int = 0) -> str:
        """Get the name of one or more data loaders, which will be prepended to all logs."""
        return self.validation_names[dataloader_idx]  # type: ignore

    def get_test_dataloader_prefix(self, dataloader_idx: int = 0) -> str:
        """Get the name of one or more data loaders, which will be prepended to all logs."""
        return self.test_names[dataloader_idx]  # type: ignore

    def load_part_of_state_dict(self, state_dict, include, exclude, load_from_string=None):
        """Load part of the state dict."""
        excluded_param_names = []
        # create dict
        dict_to_load = {}
        for k, v in state_dict.items():
            should_add = any(p in k for p in include)
            # except for if any string from exclude is present
            for e in exclude:
                if e in k:
                    excluded_param_names.append(k)
                    should_add = False
                    break
            if should_add:
                dict_to_load[k] = v

        # Restore checkpoint part into current model
        self.load_state_dict(dict_to_load, strict=False)
        if load_from_string is not None:
            logging.info(f"Model checkpoint partially restored from {load_from_string}")
            if len(excluded_param_names) > 0:
                logging.info(
                    f"The following parameters were excluded when loading from {load_from_string} : "
                    f"{excluded_param_names}"
                )
                logging.info("Make sure that this is what you wanted!")
        else:
            if len(excluded_param_names) > 0:
                logging.info(
                    f"The following parameters were excluded when loading checkpoint : {excluded_param_names}"
                )

    @rank_zero_only
    def maybe_init_from_pretrained_checkpoint(self, cfg: OmegaConf, map_location: str = "cpu"):  # noqa: MC0001
        """Initializes a given model with the parameters obtained via specific config arguments. The state dict of the
        provided model will be updated with `strict=False` setting to prevent requirement of exact model parameters
        matching.

        Initializations
        ---------------
        init_from_atommic_model : str
            Str path to a .atommic model, which will be instantiated in order to extract the state dict.
        init_from_pretrained_model : str
            Str name of a pretrained model checkpoint (obtained via cloud). The model will be downloaded (or a cached
            copy will be used), instantiated and then its state dict will be extracted.
        init_from_ptl_ckpt : str
            Str name of a Pytorch Lightning checkpoint file. It will be loaded and the state dict will extract.

        Parameters
        ----------
        cfg : OmegaConf
            The config used to instantiate the model. It needs only contain one of the above keys.
        map_location : str or torch.device()
            str or torch.device() which represents where the intermediate state dict (from the pretrained model or
            checkpoint) will be loaded.
        """
        args = ["init_from_atommic_model", "init_from_pretrained_model", "init_from_ptl_ckpt"]
        arg_matches = [(1 if arg in cfg and arg is not None else 0) for arg in args]

        if sum(arg_matches) == 0:
            # model weights do not need to be restored
            return

        if sum(arg_matches) > 1:
            raise ValueError(
                "Cannot pass more than one model initialization arguments to config!\n"
                f"Found : {[args[idx] for idx, arg_present in enumerate(arg_matches) if arg_present]}"
            )

        if "init_from_atommic_model" in cfg and cfg.init_from_atommic_model is not None:
            with open_dict(cfg):
                if isinstance(cfg.init_from_atommic_model, str):
                    model_path = cfg.init_from_atommic_model
                    # Restore model
                    restored_model = self.restore_from(
                        model_path, map_location=map_location, strict=cfg.get("init_strict", True)
                    )
                    # Restore checkpoint into current model
                    self.load_state_dict(restored_model.state_dict(), strict=False)
                    logging.info(f"Model checkpoint restored from atommic file with path : `{model_path}`")
                elif isinstance(cfg.init_from_atommic_model, (DictConfig, dict)):
                    model_load_dict = cfg.init_from_atommic_model
                    for model_load_cfg in model_load_dict.values():
                        model_path = model_load_cfg.path
                        # Restore model
                        restored_model = self.restore_from(
                            model_path, map_location=map_location, strict=cfg.get("init_strict", True)
                        )

                        include = model_load_cfg.pop("include", [""])
                        exclude = model_load_cfg.pop("exclude", [])

                        self.load_part_of_state_dict(
                            restored_model.state_dict(), include, exclude, f"atommic file with path `{model_path}`"
                        )
                else:
                    raise TypeError("Invalid type: init_from_atommic_model is not a string or a dict!")

        if "init_from_pretrained_model" in cfg and cfg.init_from_pretrained_model is not None:
            with open_dict(cfg):
                # Restore model
                if isinstance(cfg.init_from_pretrained_model, str):
                    model_name = cfg.pop("init_from_pretrained_model")

                    # Check if model is being resumed or not - only works if `Trainer` is attached to model
                    if hasattr(self, "trainer") and self.trainer is not None:
                        trainer = self.trainer
                        if (
                            hasattr(trainer, "resume_from_checkpoint")
                            and trainer.checkpoint_connector.resume_checkpoint_path is not None
                        ):
                            logging.info(
                                "Model training is being resumed via Pytorch Lightning.\n"
                                "Initialization from pretrained model (via cloud) will be skipped."
                            )
                            return

                    restored_model = self.from_pretrained(
                        model_name, map_location=map_location, strict=cfg.get("init_strict", True)
                    )

                    # Restore checkpoint into current model
                    self.load_state_dict(restored_model.state_dict(), strict=False)
                    logging.info(f"Model checkpoint restored from pretrained checkpoint with name : `{model_name}`")
                elif isinstance(cfg.init_from_pretrained_model, dict):
                    pass
                elif isinstance(cfg.init_from_pretrained_model, (DictConfig, dict)):
                    model_load_dict = cfg.init_from_pretrained_model
                    for model_load_cfg in model_load_dict.values():
                        model_name = model_load_cfg.name
                        # Restore model
                        restored_model = self.from_pretrained(
                            model_name, map_location=map_location, strict=cfg.get("init_strict", True)
                        )

                        include = model_load_cfg.pop("include", [""])
                        exclude = model_load_cfg.pop("exclude", [])

                        self.load_part_of_state_dict(
                            restored_model.state_dict(),
                            include,
                            exclude,
                            f"pretrained checkpoint with name `{model_name}`",
                        )
                else:
                    raise TypeError("Invalid type: init_from_pretrained_model is not a string or a dict!")

        if "init_from_ptl_ckpt" in cfg and cfg.init_from_ptl_ckpt is not None:
            with open_dict(cfg):
                if isinstance(cfg.init_from_ptl_ckpt, str):
                    # Restore checkpoint
                    ckpt_path = cfg.pop("init_from_ptl_ckpt")
                    ckpt = torch.load(ckpt_path, map_location=map_location)

                    # Restore checkpoint into current model
                    self.load_state_dict(ckpt["state_dict"], strict=False)
                    logging.info(
                        f"Model checkpoint restored from pytorch lightning checkpoint with path : `{ckpt_path}`"
                    )
                elif isinstance(cfg.init_from_ptl_ckpt, (DictConfig, dict)):
                    model_load_dict = cfg.init_from_ptl_ckpt
                    for model_load_cfg in model_load_dict.values():
                        ckpt_path = model_load_cfg.path
                        # Restore model
                        ckpt = torch.load(ckpt_path, map_location=map_location)

                        include = model_load_cfg.pop("include", [""])
                        exclude = model_load_cfg.pop("exclude", [])

                        self.load_part_of_state_dict(
                            ckpt["state_dict"], include, exclude, f"atommic file with path `{ckpt_path}`"
                        )
                else:
                    raise TypeError("Invalid type: init_from_ptl_ckpt is not a string or a dict!")

    def teardown(self, stage: str):
        """Called at the end of fit and test."""
        if stage == "fit" and "PL_TRAINER_GPUS" in os.environ:
            os.environ.pop("PL_TRAINER_GPUS")

        super().teardown(stage)

    @classmethod
    def extract_state_dict_from(
        cls,
        restore_path: str,
        save_dir: str,
        split_by_module: bool = False,
        save_restore_connector: SaveRestoreConnector = None,
    ):
        """Extract the state dict(s) from a provided .atommic tarfile and save it to a directory.

        Parameters
        ----------
        restore_path : str
            Path to .atommic file from which state dict(s) should be extracted.
        save_dir : str
            Directory in which the saved state dict(s) should be stored.
        split_by_module : bool, optional
            Bool flag, which determines whether the output checkpoint should be for the entire Model, or
            the individual module's that comprise the Model. Default is ``False``.
        save_restore_connector : SaveRestoreConnector, optional
            Can be overridden to add custom save and restore logic. Default is ``None``.

        Example
        -------
        To convert the .atommic tarfile into a single Model level PyTorch checkpoint

        .. code-block::

            state_dict = atommic.collections.asr.models.EncDecCTCModel.extract_state_dict_from('asr.atommic', \
            './asr_ckpts')

        To restore a model from a Model level checkpoint

        .. code-block::

            model = atommic.collections.asr.models.EncDecCTCModel(cfg)  # or any other method of restoration
            model.load_state_dict(torch.load("./asr_ckpts/model_weights.ckpt"))

        To convert the .atommic tarfile into multiple Module level PyTorch checkpoints

        .. code-block::

            state_dict = atommic.collections.asr.models.EncDecCTCModel.extract_state_dict_from('asr.atommic', \
            './asr_ckpts', split_by_module=True)

        To restore a module from a Module level checkpoint

        .. code-block::

            model = atommic.collections.asr.models.EncDecCTCModel(cfg)  # or any other method of restoration
            # load the individual components
            model.preprocessor.load_state_dict(torch.load("./asr_ckpts/preprocessor.ckpt"))
            model.encoder.load_state_dict(torch.load("./asr_ckpts/encoder.ckpt"))
            model.decoder.load_state_dict(torch.load("./asr_ckpts/decoder.ckpt"))

        Returns
        -------
        The state dict that was loaded from the original .atommic checkpoint.
        """
        if save_restore_connector is None:
            save_restore_connector = SaveRestoreConnector()

        if not path.exists(restore_path):
            raise FileExistsError(f"Can't find {restore_path}")

        cls.update_save_restore_connector(save_restore_connector)
        return cls._save_restore_connector.extract_state_dict_from(restore_path, save_dir, split_by_module)

    def prepare_test(self, trainer: "Trainer") -> bool:
        """Helper method to check whether the model can safely be tested on a dataset after training (or loading a
        checkpoint).

        .. code-block::

            trainer = Trainer()
            if model.prepare_test(trainer):
                trainer.test(model)

        Returns
        -------
        Bool which declares the model safe to test. Provides warnings if it has to return False to guide the user.
        """
        if not hasattr(self._cfg, "test_ds"):
            logging.info("No `test_ds` config found within the manifest.")
            return False

        if trainer is not None and trainer.num_devices > 1:
            # Replace ddp multi-gpu until PTL has a fix
            DDP_WARN = """\n\nDuring testing, it is currently advisable to construct a new Trainer "
                    "with single GPU and no DDP to obtain accurate results.
                    "Following pattern should be used: "
                    "trainer = Trainer(devices=1, accelerator='gpu')
                    "if model.prepare_test(trainer):"
                    "  trainer.test(model)\n\n"""

            logging.warning(DDP_WARN)
            return False

        # Assign trainer to the model
        self.set_trainer(trainer)
        return True

    def set_trainer(self, trainer: Trainer):
        """Set an instance of Trainer object."""
        self.trainer = trainer
        self._trainer = trainer
        self.set_world_size(self._trainer)

    def set_world_size(self, trainer: Trainer):
        """Determines the world size from the PyTorch Lightning Trainer and then updates AppState."""
        self.world_size = 1

        if trainer is not None:
            if isinstance(trainer, Trainer):
                if trainer.num_devices and trainer.num_nodes:
                    self.world_size = trainer.num_devices * trainer.num_nodes
            else:
                logging.warning("World size can only be set by PyTorch Lightning Trainer.")
        app_state = AppState()
        app_state.world_size = self.world_size

    def summarize(self, max_depth: int = 1) -> model_summary.ModelSummary:
        """Summarize this LightningModule."""
        return model_summary.summarize(self, max_depth=max_depth)

    def _update_dataset_config(self, dataset_name: str, config: Optional[Union[DictConfig, Dict]]):
        """Update the config (if not None) of the dataset by given name. Preserves said config after updating.

        Parameters
        ----------
        dataset_name : str
            Name of the dataset to update. Can be one of `train`, `validation` and `test`.
        config : Optional[Union[DictConfig, Dict]]
            Config to update the dataset with. If None is passed, this method simply returns. If dict is passed, it is
            cast into a DictConfig. The internal config is updated with the passed config.
        """
        if hasattr(self, "_multi_dataset_mode") and self._multi_dataset_mode is True:
            return

        if config is not None:
            if not isinstance(config, DictConfig):
                config = OmegaConf.create(config)

            if dataset_name in {"train", "validation", "test"}:
                OmegaConf.set_struct(self.cfg, False)

                key_name = f"{dataset_name}_ds"
                self.cfg[key_name] = config

                OmegaConf.set_struct(self.cfg, True)

                # Update hyperparameters by calling property setter
                self.cfg = self._cfg
            else:
                raise ValueError("`dataset_name` when updating config must be one of [train, validation, test]")

    @property
    def num_weights(self):
        """Utility property that returns the total number of parameters of the Model."""
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    @property
    def cfg(self):
        """Property that holds the finalized internal config of the model.

        .. note::
            Changes to this config are not reflected in the state of the model.
            Please create a new model using an updated config to properly update the model.
        """
        return self._cfg

    @cfg.setter
    def cfg(self, cfg):
        """Property that holds the finalized internal config of the model.

        .. note::
            Changes to this config are not reflected in the state of the model.
            Please create a new model using an updated config to properly update the model.
        """
        self._cfg = cfg
        self._set_hparams(OmegaConf.create({"cfg": self._cfg}))

        # TODO: Remove this when we have a better way to handle this
        if hasattr(self, "_hparams_initial") and "cfg" in self._hparams_initial:
            self._hparams_initial["cfg"] = OmegaConf.to_object(self._cfg)

    @staticmethod
    def _is_model_being_restored() -> bool:
        """Checks if the model is being restored from a checkpoint."""
        app_state = AppState()
        return app_state.is_model_being_restored

    @staticmethod
    def _set_model_restore_state(is_being_restored: bool, folder: str = None):
        """Sets the state of the model to be restored."""
        app_state = AppState()
        app_state.is_model_being_restored = is_being_restored
        app_state.atommic_file_folder = folder  # type: ignore

    def _set_model_guid(self):
        """Sets the model guid."""
        if not hasattr(self, "model_guid"):
            appstate = AppState()

            # Generate a unique uuid for the instance
            # also determine if the model is being restored or not, and preserve the path
            self.model_guid = str(uuid.uuid4())
            if self._is_model_being_restored():
                restore_path = appstate.model_restore_path
            else:
                restore_path = None

            appstate.register_model_guid(self.model_guid, restoration_path=restore_path)

    @classmethod
    def update_save_restore_connector(cls, save_restore_connector):
        """Update the save_restore_connector of the model."""
        if hasattr(cls, "_save_restore_connector"):
            cls._save_restore_connector = save_restore_connector
        else:
            setattr(cls, "_save_restore_connector", save_restore_connector)

    def _setup_nsys_profiling(self):
        """Enables nsys profiling To use, add the following options to the model config:
            nsys_profile: False
                start_step: 10  # Global batch to start profiling
                end_step: 10 # Global batch to end profiling
                ranks: [0] # Global rank IDs to profile
                gen_shape: False # Generate model and kernel details including input shapes

        And then wrap the model training script with:
            nsys profile -s none -o <profile filepath>  -t cuda,nvtx --force-overwrite true \
            --capture-range=cudaProfilerApi --capture-range-end=stop python ./examples/...
        See more options at: https://docs.nvidia.com/nsight-systems/UserGuide/index.html#cli-profiling
        """
        if self.cfg.get("nsys_profile", None) is not None and self.cfg.nsys_profile.get("enabled", False):
            # Nsys profiling options
            self._nsys_profile_enabled = True
            self._nsys_profile_start_step = self.cfg.nsys_profile.get("start_step", 0)
            self._nsys_profile_end_step = self.cfg.nsys_profile.get("end_step", 0)
            self._nsys_profile_ranks = self.cfg.nsys_profile.get("ranks", [0])
            self._nsys_profile_gen_shape = self.cfg.nsys_profile.get("gen_shape", False)

            if isinstance(self._nsys_profile_start_step, int):
                logging.info(f"Nsys profiling setup with start_step: {self._nsys_profile_start_step}")
            else:
                raise ValueError(f"Nsys start_step must be of type int. Found: {type(self._nsys_profile_start_step)}")

            if isinstance(self._nsys_profile_end_step, int):
                logging.info(f"Nsys profiling setup with end_step: {self._nsys_profile_end_step}")
            else:
                raise ValueError(f"Nsys end_step must be of type int. Found: {type(self._nsys_profile_end_step)}")

            if self._nsys_profile_end_step >= self._nsys_profile_start_step:
                pass
            else:
                raise ValueError("Nsys end_step must be greater than or equal to nsys start_step")

    def on_train_start(self):
        """PyTorch Lightning hook:
        https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#on-train-start
        We use it here to copy the relevant config for dynamic freezing.
        """

        # dynamic freezing
        # should fire only once, on the very first batch of training and never again
        if not hasattr(self, "_freeze_cfg"):
            if (
                hasattr(self.cfg, "freeze_updates")
                and self.cfg.freeze_updates is not None
                and self.cfg.freeze_updates.get("enabled", False)
            ):
                setattr(self, "_freeze_cfg", OmegaConf.to_container(self.cfg.freeze_updates))
                self._freeze_cfg["is_frozen"] = {k: False for k in self._freeze_cfg["modules"].keys()}
            else:
                setattr(self, "_freeze_cfg", None)

    def on_train_batch_start(self, batch: Any, batch_idx: int, unused: int = 0):  # pylint: disable=unused-argument
        """PyTorch Lightning hook:
        https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#on-train-batch-start
        We use it here to enable nsys profiling and dynamic freezing.
        """

        # nsys profiling
        if self.device.type == "cuda":
            if hasattr(self, "_nsys_profile_enabled"):
                if self._nsys_profile_enabled:
                    if batch_idx == self._nsys_profile_start_step and get_rank() in self._nsys_profile_ranks:
                        logging.info("====== Start nsys profiling ======")
                        torch.cuda.cudart().cudaProfilerStart()
                        if self._nsys_profile_gen_shape:
                            torch.autograd.profiler.emit_nvtx(  # pylint: disable=unnecessary-dunder-call
                                record_shapes=True
                            ).__enter__()

        # dynamic freezing
        if hasattr(self, "_freeze_cfg") and self._freeze_cfg is not None:
            if self.training and hasattr(self, "trainer") and self.trainer is not None:
                num_updates = self.trainer.global_step + 1

                for ml, m_steps in self._freeze_cfg["modules"].items():
                    # we could do hasattr check here, but it's too expensive for each step consequently you'll throw an
                    # error if the module name doesn't exist or was spelled wrong in the config.yaml
                    if isinstance(m_steps, list):
                        assert len(m_steps) == 2, "freeze_updates modules list cannot have more than two elements"
                        should_freeze = (num_updates >= m_steps[0]) and (num_updates <= m_steps[1] or m_steps[1] == -1)
                    else:
                        should_freeze = num_updates <= m_steps or m_steps == -1
                    if should_freeze and not self._freeze_cfg["is_frozen"][ml]:
                        getattr(self, ml).freeze()
                        getattr(self, ml).train()
                        self._freeze_cfg["is_frozen"][ml] = True
                    elif not should_freeze and self._freeze_cfg["is_frozen"][ml]:
                        getattr(self, ml).unfreeze()
                        self._freeze_cfg["is_frozen"][ml] = False

    def on_train_batch_end(
        self, outputs, batch: Any, batch_idx: int, unused: int = 0  # pylint: disable=unused-argument
    ) -> None:
        """PyTorch Lightning hook:
        https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#on-train-batch-end
        We use it here to enable nsys profiling.
        """
        if (
            self.device.type == "cuda"
            and hasattr(self, "_nsys_profile_enabled")
            and self._nsys_profile_enabled
            and batch_idx == self._nsys_profile_end_step
            and get_rank() in self._nsys_profile_ranks
        ):
            logging.info("====== End nsys profiling ======")
            torch.cuda.cudart().cudaProfilerStop()

    def on_train_end(self):
        """PyTorch Lightning hook:
        https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#on-train-end
        We use it here to clean up the dynamic freezing config.
        """
        # dynamic freezing cleanup
        if hasattr(self, "_freeze_cfg"):
            delattr(self, "_freeze_cfg")

    def cuda(self, device=None):
        """PTL is overriding this method and changing the pytorch behavior of a module. The PTL LightingModule
        override will move the module to device 0 if device is None.

            See the PTL method here:
            https://github.com/Lightning-AI/lightning/blob/master/src/pytorch_lightning/core/mixins/
            device_dtype_mixin.py#L113

            Here we are overriding this to maintain the default Pytorch nn.module behavior:
            https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/module.py#L728

        Moves all model parameters and buffers to the GPU.

        This also makes associated parameters and buffers different objects. So
        it should be called before constructing optimizer if the module will
        live on GPU while being optimized.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device, optional
            The desired device of the parameters and buffers in this module.
        """
        if device is None:
            device = torch.device("cuda", torch.cuda.current_device())
        elif isinstance(device, int):
            device = torch.device("cuda", index=device)
        return super().cuda(device=device)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/nn/crossdomain/crossdomain.py

from typing import Optional, Tuple, Union

import torch
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul


class MultiCoil(nn.Module):
    """Makes the forward pass of multi-coil data of shape (N, N_coils, H, W, C) to a model.

    If coil_to_batch is set to True, coil dimension is moved to the batch dimension. Otherwise, it passes to the model
    each coil-data individually.
    """

    def __init__(self, model: nn.Module, coil_dim: int = 1, coil_to_batch: bool = False):
        """Inits :class:`MultiCoil`.

        Parameters
        ----------
        model : torch.nn.Module
            Any nn.Module that takes as input with 4D data (N, H, W, C). Typically, a convolutional-like model.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        coil_to_batch : bool, optional
            If True batch and coil dimensions are merged when forwarded by the model and unmerged when outputted.
            Otherwise, input is forwarded to the model per coil. Default is ``False``.
        """
        super().__init__()
        self.model = model
        self.coil_to_batch = coil_to_batch
        self.coil_dim = coil_dim

    def _compute_model_per_coil(self, data: torch.Tensor) -> torch.Tensor:
        """Computes the model per coil."""
        output = []
        for idx in range(data.size(self.coil_dim)):
            subselected_data = data.select(self.coil_dim, idx)
            if subselected_data.shape[-1] == 2 and subselected_data.dim() == 4:
                output.append(self.model(subselected_data.permute(0, 3, 1, 2)))
            else:
                output.append(self.model(subselected_data.unsqueeze(self.coil_dim)).squeeze(self.coil_dim))
        output = torch.stack(output, dim=self.coil_dim)
        return output

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`MultiCoil`.

        Parameters
        ----------
        x : torch.Tensor
            Input data of shape (N, N_coils, H, W, C).

        Returns
        -------
        torch.Tensor
            Output data of shape (N, N_coils, H, W, C).
        """
        if self.coil_to_batch:
            x = x.clone()

            batch, coil, channels, height, width = x.size()
            x = x.reshape(batch * coil, channels, height, width).contiguous()
            x = self.model(x).permute(0, 2, 3, 1)
            x = x.reshape(batch, coil, height, width, -1).permute(0, 1, 4, 2, 3)
        else:
            x = self._compute_model_per_coil(x).contiguous()

        return x


class CrossDomainNetwork(nn.Module):
    """Based on KIKINet implementation. Modified to work with multi-coil k-space data, as presented in [Taejoon2018]_.

    This performs optimisation in both, k-space ("K") and image ("I") domains according to domain_sequence.

    References
    ----------
    .. [Taejoon2018] Eo, Taejoon, et al. “KIKI-Net: Cross-Domain Convolutional Neural Networks for Reconstructing
        Undersampled Magnetic Resonance Images.” Magnetic Resonance in Medicine, vol. 80, no. 5, Nov. 2018, pp.
        2188–201. PubMed,  https://doi.org/10.1002/mrm.27201.
    """

    def __init__(
        self,
        image_model_list: nn.Module,
        kspace_model_list: Optional[Union[nn.Module, None]] = None,
        domain_sequence: str = "KIKI",
        image_buffer_size: int = 1,
        kspace_buffer_size: int = 1,
        normalize_image: bool = False,  # pylint: disable=unused-argument
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
    ):
        """Inits :class:`CrossDomainNetwork`.

        Parameters
        ----------
        image_model_list : torch.nn.Module
            Image domain model list.
        kspace_model_list : torch.nn.Module, optional
            K-space domain model list. If set to None, a correction step is applied. Default is ``None``.
        domain_sequence : str, optional
            Domain sequence. Default is ``"KIKI"``.
        image_buffer_size : int, optional
            Image buffer size. Default is ``1``.
        kspace_buffer_size : int, optional
            K-space buffer size. Default is ``1``.
        normalize_image : bool, optional
            Whether to normalize the image. Default is ``False``.
        fft_centered : bool, optional
            Whether to use centered FFT. Default is ``False``.
        fft_normalization : str, optional
            Whether to normalize the FFT. Default is ``"backward"``.
        spatial_dims : Tuple[int, int], optional
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        """
        super().__init__()

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        domain_sequence = list(domain_sequence.strip())  # type: ignore
        if not set(domain_sequence).issubset({"K", "I"}):
            raise ValueError(f"Invalid domain sequence. Got {domain_sequence}. Should only contain 'K' and 'I'.")
        if kspace_model_list is not None and len(kspace_model_list) != domain_sequence.count("K"):
            raise ValueError("K-space domain steps do not match k-space model list length.")
        if len(image_model_list) != domain_sequence.count("I"):
            raise ValueError("Image domain steps do not match image model list length.")

        self.domain_sequence = domain_sequence
        self.kspace_model_list = kspace_model_list
        self.kspace_buffer_size = kspace_buffer_size
        self.image_model_list = image_model_list
        self.image_buffer_size = image_buffer_size

    def kspace_correction(
        self,
        block_idx: int,
        image_buffer: torch.Tensor,
        kspace_buffer: torch.Tensor,
        sampling_mask: torch.Tensor,
        sensitivity_map: torch.Tensor,
        masked_kspace: torch.Tensor,
    ) -> torch.Tensor:
        """Performs k-space correction.

        Parameters
        ----------
        block_idx : int
            Block index.
        image_buffer : torch.Tensor
            Image buffer.
        kspace_buffer : torch.Tensor
            K-space buffer.
        sampling_mask : torch.Tensor
            Subsampling mask.
        sensitivity_map : torch.Tensor
            Coil sensitivity maps.
        masked_kspace : torch.Tensor
            Subsampled k-space.

        Returns
        -------
        torch.Tensor
            K-space buffer.
        """
        forward_buffer = [
            self._forward_operator(image.clone(), sampling_mask, sensitivity_map)
            for image in torch.split(image_buffer, 2, -1)
        ]
        forward_buffer = torch.cat(forward_buffer, -1)

        kspace_buffer = torch.cat([kspace_buffer, forward_buffer, masked_kspace], -1)

        if self.kspace_model_list is not None:
            kspace_buffer = self.kspace_model_list[block_idx](kspace_buffer.permute(0, 1, 4, 2, 3)).permute(
                0, 1, 3, 4, 2
            )
        else:
            kspace_buffer = kspace_buffer[..., :2] - kspace_buffer[..., 2:4]

        return kspace_buffer

    def image_correction(
        self,
        block_idx: int,
        image_buffer: torch.Tensor,
        kspace_buffer: torch.Tensor,
        sampling_mask: torch.Tensor,
        sensitivity_map: torch.Tensor,
    ) -> torch.Tensor:
        """Performs image space correction.

        Parameters
        ----------
        block_idx : int
            Block index.
        image_buffer : torch.Tensor
            Image buffer.
        kspace_buffer : torch.Tensor
            K-space buffer.
        sampling_mask : torch.Tensor
            Subsampling mask.
        sensitivity_map : torch.Tensor
            Coil sensitivity maps.

        Returns
        -------
        torch.Tensor
            Image buffer.
        """
        backward_buffer = [
            self._backward_operator(kspace.clone(), sampling_mask, sensitivity_map)
            for kspace in torch.split(kspace_buffer, 2, -1)
        ]
        backward_buffer = torch.cat(backward_buffer, -1)
        image_buffer = torch.cat([image_buffer, backward_buffer], -1).permute(0, 3, 1, 2)
        image_buffer = self.image_model_list[block_idx](image_buffer).permute(0, 2, 3, 1)
        return image_buffer

    def _forward_operator(
        self,
        image: torch.Tensor,
        sampling_mask: torch.Tensor,
        sensitivity_map: torch.Tensor,
    ) -> torch.Tensor:
        """Custom forward operator for the cross-domain correction.

        Parameters
        ----------
        image : torch.Tensor
            Image space. Shape [batch, coils, height, width, 2].
        sampling_mask : torch.Tensor
            Subsampling mask. Shape [batch, 1, height, width, 1].
        sensitivity_map : torch.Tensor
            Coil sensitivity maps. Shape [batch, coils, height, width, 2].

        Returns
        -------
        torch.Tensor
            K-space prediction. Shape [batch, coils, height, width, 2].
        """
        return torch.where(
            sampling_mask == 0,
            torch.tensor([0.0], dtype=image.dtype).to(image.device),
            fft2(
                complex_mul(image.unsqueeze(self.coil_dim), sensitivity_map),
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            ).type(image.type()),
        )

    def _backward_operator(
        self,
        kspace: torch.Tensor,
        sampling_mask: torch.Tensor,
        sensitivity_map: torch.Tensor,
    ) -> torch.Tensor:
        """Custom backward operator for the cross-domain correction.

        Parameters
        ----------
        kspae : torch.Tensor
            K-space. Shape [batch, coils, height, width, 2].
        sampling_mask : torch.Tensor
            Subsampling mask. Shape [batch, 1, height, width, 1].
        sensitivity_map : torch.Tensor
            Coil sensitivity maps. Shape [batch, coils, height, width, 2].

        Returns
        -------
        torch.Tensor
            Image space prediction. Shape [batch, coils, height, width, 2].
        """
        kspace = torch.where(sampling_mask == 0, torch.tensor([0.0], dtype=kspace.dtype).to(kspace.device), kspace)
        return (
            complex_mul(
                ifft2(
                    kspace.float(),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                complex_conj(sensitivity_map),
            )
            .sum(self.coil_dim)
            .type(kspace.type())
        )

    @staticmethod
    def crop_to_shape(x: torch.Tensor, shape: tuple) -> torch.Tensor:
        r"""Crops ``x`` to specified shape.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape ('\'*, H, W).
        shape : tuple
            Crop shape corresponding to H, W.

        Returns
        -------
        torch.Tensor
            Cropped tensor.
        """
        h, w = x.shape[1:3]
        if h > shape[0]:
            x = x[:, : shape[0], :, :]
        if w > shape[1]:
            x = x[:, :, : shape[1], :]
        return x

    def forward(
        self,
        masked_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        sampling_mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of :class:`CrossDomainNetwork`.

        Parameters
        ----------
        masked_kspace : torch.Tensor
            Subsampled k-space. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        sampling_mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]

        Returns
        -------
        torch.Tensor
            Reconstructed image. Shape [batch_size, n_x, n_y, 2]
        """
        input_image = self._backward_operator(masked_kspace, sampling_mask, sensitivity_maps)

        image_buffer = torch.cat([input_image] * self.image_buffer_size, -1).to(masked_kspace.device)
        kspace_buffer = torch.cat([masked_kspace] * self.kspace_buffer_size, -1).to(masked_kspace.device)

        kspace_block_idx, image_block_idx = 0, 0
        for block_domain in self.domain_sequence:
            if block_domain == "K":
                kspace_buffer = self.kspace_correction(
                    kspace_block_idx, image_buffer, kspace_buffer, sampling_mask, sensitivity_maps, masked_kspace
                )
                kspace_block_idx = kspace_block_idx + 1
            else:
                image_buffer = self.image_correction(
                    image_block_idx, image_buffer, kspace_buffer, sampling_mask, sensitivity_maps
                )
                image_buffer = self.crop_to_shape(image_buffer, sensitivity_maps.shape[2:4])
                image_block_idx = image_block_idx + 1

        return image_buffer[..., :2]



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.core.classes.common import typecheck

__all__ = ["ZF"]


class ZF(BaseMRIReconstructionModel):
    """Zero-Filled reconstruction using either root-sum-of-squares (RSS) or SENSE (SENSitivity Encoding, as presented
    in [Pruessmann1999]_).

    References
    ----------
    .. [Pruessmann1999] Pruessmann KP, Weiger M, Scheidegger MB, Boesiger P. SENSE: Sensitivity encoding for fast MRI.
        Magn Reson Med 1999; 42:952-962.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`ZF`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,  # pylint: disable=unused-argument
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`ZF`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        return check_stacked_complex(
            coil_combination_method(
                ifft2(y, self.fft_centered, self.fft_normalization, self.spatial_dims),
                sensitivity_maps,
                self.coil_combination_method.upper(),
                self.coil_dim,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.reconstruction.data.mri_reconstruction_loader import (  # noqa: F401
    CC359ReconstructionMRIDataset,
    ReconstructionMRIDataset,
    SKMTEAReconstructionMRIDataset,
    StanfordKneesReconstructionMRIDataset,
)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from https://github.com/dfdazac/wassdistance/blob/master/layers.py

import torch

from atommic.core.classes.loss import Loss


class SinkhornDistance(Loss):
    r"""Given two empirical measures each with :math:`P_1` locations :math:`x\in\mathbb{R}^{D_1}` and :math:`P_2`
    locations :math:`y\in\mathbb{R}^{D_2}`, outputs an approximation of the regularized OT cost for point clouds.
    """

    def __init__(self, eps=0.1, max_iter=100, reduction="mean"):
        """Inits :class:`SinkhornDistance`.

        Parameters
        ----------
        eps : float
            Regularization coefficient. Default is ``0.1``.
        max_iter : int
            Maximum number of Sinkhorn iterations. Default is ``100``.
        reduction : string, optional
            Specifies the reduction to apply to the output:
            'none' | 'mean' | 'sum'. 'none': no reduction will be applied.
            Default is ``mean``.
        """
        super().__init__()
        self.eps = torch.tensor([eps])
        self.max_iter = max_iter
        self.reduction = reduction

    def forward(self, x, y):
        r"""Forward pass of the Sinkhorn algorithm.

        Parameters
        ----------
        x : torch.Tensor
            :math:`(N, P_1, D_1)`
        y : torch.Tensor
            :math:`(N, P_2, D_2)`

        Returns
        -------
        torch.Tensor
            The Sinkhorn distance between the two point clouds. Output shape :math:`(N)` or :math:`()`, depending on
            `reduction`
        """
        self.eps = self.eps.clone().to(x.device)

        # The Sinkhorn algorithm takes as input three variables :
        C = self._cost_matrix(x, y)  # Wasserstein cost function
        x_points = x.shape[-2]
        y_points = y.shape[-2]
        if x.dim() == 2:
            batch_size = 1
        else:
            batch_size = x.shape[0]

        # both marginals are fixed with equal weights
        mu = (
            torch.empty(batch_size, x_points, dtype=torch.float, requires_grad=False)
            .fill_(1.0 / x_points)
            .squeeze()
            .to(x.device)
        )
        nu = (
            torch.empty(batch_size, y_points, dtype=torch.float, requires_grad=False)
            .fill_(1.0 / y_points)
            .squeeze()
            .to(x.device)
        )

        u = torch.zeros_like(mu)
        v = torch.zeros_like(nu)
        # To check if algorithm terminates because of threshold or max iterations reached
        actual_nits = 0
        # Stopping criterion
        thresh = 1e-1

        # Sinkhorn iterations
        for _ in range(self.max_iter):
            u1 = u  # useful to check the update
            u = self.eps * (torch.log(mu + 1e-8) - torch.logsumexp(self.M(C, u, v), dim=-1)) + u
            v = self.eps * (torch.log(nu + 1e-8) - torch.logsumexp(self.M(C, u, v).transpose(-2, -1), dim=-1)) + v
            err = (u - u1).abs().sum(-1).mean()

            actual_nits = actual_nits + 1
            if err.item() < thresh:
                break

        U, V = u, v
        # Transport plan pi = diag(a)*K*diag(b)
        pi = torch.exp(self.M(C, U, V))
        # Sinkhorn distance
        cost = torch.sum(pi * C, dim=(-2, -1))

        if self.reduction == "mean":
            cost = cost.mean()
        elif self.reduction == "sum":
            cost = cost.sum()

        return cost  # , pi, C

    def M(self, C, u, v):
        r"""Modified cost for logarithmic updates $M_{ij} = (-c_{ij} + u_i + v_j) / \epsilon$"""
        return (-C + u.unsqueeze(-1) + v.unsqueeze(-2)) / self.eps

    @staticmethod
    def _cost_matrix(x, y, p=2):
        r"""Returns the matrix of $|x_i-y_j|^p$."""
        x_col = x.unsqueeze(-2)
        y_lin = y.unsqueeze(-3)
        C = torch.sum((torch.abs(x_col - y_lin)) ** p, -1)
        return C

    @staticmethod
    def ave(u, u1, tau):
        r"""Barycenter subroutine, used by kinetic acceleration through extrapolation."""
        return tau * u + (1 - tau) * u1



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import List, Tuple

import torch
from torch import Tensor, nn

from atommic.collections.reconstruction.nn.unet_base.unet_block import ConvBlock, TransposeConvBlock


class DC(nn.Module):
    """IDSLR Data consistency block, as presented."""

    def __init__(self):
        """Inits :class:`DC`."""
        super().__init__()
        self.dc_weight = nn.Parameter(torch.ones(1))

    def forward(self, prediction_kspace: Tensor, reference_kspace: Tensor, mask: Tensor) -> Tensor:
        """Forward pass of :class:`DC`.

        Parameters
        ----------
        prediction_kspace : torch.Tensor
            Prediction k-space. Shape: (batch, channels, height, width, complex)
        reference_kspace : torch.Tensor
            Reference k-space. Shape: (batch, channels, height, width, complex)
        mask : torch.Tensor
            Subsampling mask. Shape: (batch, channels, height, width, 1)

        Returns
        -------
        torch.Tensor
            Data consistency k-space. Shape: (batch, channels, height, width, complex)
        """
        return torch.div(
            torch.view_as_complex(reference_kspace) + self.dc_weight * torch.view_as_complex(prediction_kspace),
            mask.squeeze(-1) + torch.complex(self.dc_weight, torch.zeros_like(self.dc_weight)),
        )


class UnetEncoder(nn.Module):
    """UNet Encoder block, according to the implementation of the NormUnet."""

    def __init__(
        self,
        chans: int,
        num_pools: int,
        in_chans: int = 2,
        drop_prob: float = 0.0,
        normalize: bool = True,
        padding: bool = True,
        padding_size: int = 15,
        norm_groups: int = 2,
    ):
        """Inits :class:`UnetEncoder`.

        Parameters
        ----------
        chans : int
            Number of channels in the first layer.
        num_pools : int
            Number of down-sampling layers.
        in_chans : int, optional
            Number of input channels. Default is ``2``.
        drop_prob : float, optional
            Dropout probability. Default is ``0.0``.
        normalize : bool, optional
            Whether to normalize the input. Default is ``True``.
        padding : bool, optional
            Whether to pad the input. Default is ``True``.
        padding_size : int, optional
            Padding size. Default is ``15``.
        norm_groups : int, optional
            Number of groups for group normalization. Default is ``2``.
        """
        super().__init__()

        self.in_chans = in_chans
        self.chans = chans
        self.num_pools = num_pools
        self.drop_prob = drop_prob
        self.normalize = normalize
        self.padding = padding
        self.padding_size = padding_size
        self.norm_groups = norm_groups

        self.down_sample_layers = torch.nn.ModuleList([ConvBlock(in_chans, chans, drop_prob)])
        ch = chans
        for _ in range(num_pools - 1):
            self.down_sample_layers.append(ConvBlock(ch, ch * 2, drop_prob))
            ch = ch * 2
        self.conv = ConvBlock(ch, ch * 2, drop_prob)

    @staticmethod
    def complex_to_chan_dim(x: torch.Tensor) -> torch.Tensor:
        """Converts the last dimension of the input to complex."""
        b, c, h, w, two = x.shape
        if two != 2:
            raise AssertionError
        return x.permute(0, 4, 1, 2, 3).reshape(b, 2 * c, h, w)

    def norm(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Normalizes the input."""
        # group norm
        b, c, h, w = x.shape
        x = x.reshape(b, self.norm_groups, -1)
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)

        x = (x - mean) / std
        x = x.reshape(b, c, h, w)
        return x, mean, std

    def pad(self, x: torch.Tensor) -> Tuple[torch.Tensor, Tuple[List[int], List[int], int, int]]:
        """Pads the input with zeros to make it square."""
        _, _, h, w = x.shape
        w_mult = ((w - 1) | self.padding_size) + 1
        h_mult = ((h - 1) | self.padding_size) + 1
        w_pad = [math.floor((w_mult - w) / 2), math.ceil((w_mult - w) / 2)]
        h_pad = [math.floor((h_mult - h) / 2), math.ceil((h_mult - h) / 2)]
        x = torch.nn.functional.pad(x, w_pad + h_pad)
        return x, (h_pad, w_pad, h_mult, w_mult)

    def forward(
        self, x: torch.Tensor
    ) -> Tuple[List[torch.Tensor], bool, Tuple[List[int], List[int], int, int], torch.Tensor, torch.Tensor]:
        """Forward pass of :class:`UnetEncoder`.

        Parameters
        ----------
        x : torch.Tensor
            Input data. Shape: (batch, channels, height, width, complex)

        Returns
        -------
        List[torch.Tensor]
            List of down-sampled layers.
        bool
            Whether the input was complex.
        Tuple[List[int], List[int], int, int]
            Padding sizes.
        torch.Tensor
            Mean of the input.
        torch.Tensor
            Standard deviation of the input.
        """
        iscomplex = False
        if x.shape[-1] == 2:
            x = self.complex_to_chan_dim(x)
            iscomplex = True

        if self.normalize:
            x, mean, std = self.norm(x)
        else:
            mean = 1.0
            std = 1.0

        if self.padding:
            x, pad_sizes = self.pad(x)
        else:
            pad_sizes = None

        stack = []
        output = x

        # apply down-sampling layers
        for layer in self.down_sample_layers:
            output = layer(output)
            stack.append(output)
            output = torch.nn.functional.avg_pool2d(output, kernel_size=2, stride=2, padding=0)

        output = self.conv(output)
        stack.append(output)

        return stack, iscomplex, pad_sizes, mean, std  # type: ignore


class UnetDecoder(nn.Module):
    """UNet Encoder block, according to the implementation of the NormUnet."""

    def __init__(
        self,
        chans: int,
        num_pools: int,
        out_chans: int = 2,
        drop_prob: float = 0.0,
        normalize: bool = True,
        padding: bool = True,
        padding_size: int = 15,
        norm_groups: int = 2,
    ):
        """Inits :class:`UnetDecoder`.

        Parameters
        ----------
        chans : int
            Number of channels in the first layer.
        num_pools : int
            Number of down-sampling layers.
        in_chans : int, optional
            Number of input channels. Default is ``2``.
        drop_prob : float, optional
            Dropout probability. Default is ``0.0``.
        normalize : bool, optional
            Whether to normalize the input. Default is ``True``.
        padding : bool, optional
            Whether to pad the input. Default is ``True``.
        padding_size : int, optional
            Padding size. Default is ``15``.
        norm_groups : int, optional
            Number of groups for group normalization. Default is ``2``.
        """
        super().__init__()

        self.out_chans = out_chans
        self.chans = chans
        self.num_pools = num_pools
        self.drop_prob = drop_prob
        self.normalize = normalize
        self.padding = padding
        self.padding_size = padding_size
        self.norm_groups = norm_groups

        ch = chans * (2 ** (num_pools - 1))
        self.up_conv = torch.nn.ModuleList()
        self.up_transpose_conv = torch.nn.ModuleList()
        for _ in range(num_pools - 1):
            self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
            self.up_conv.append(ConvBlock(ch * 2, ch, drop_prob))
            ch = ch // 2

        self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
        self.up_conv.append(
            torch.nn.Sequential(
                ConvBlock(ch * 2, ch, drop_prob),
                torch.nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        )

    @staticmethod
    def chan_complex_to_last_dim(x: torch.Tensor) -> torch.Tensor:
        """Converts the last dimension of the input to complex."""
        b, c2, h, w = x.shape
        if c2 % 2 != 0:
            raise AssertionError
        c = torch.div(c2, 2, rounding_mode="trunc")
        return x.view(b, 2, c, h, w).permute(0, 2, 3, 4, 1).contiguous()

    @staticmethod
    def unpad(x: torch.Tensor, h_pad: List[int], w_pad: List[int], h_mult: int, w_mult: int) -> torch.Tensor:
        """Unpads the input."""
        return x[..., h_pad[0] : h_mult - h_pad[1], w_pad[0] : w_mult - w_pad[1]]

    def unnorm(self, x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        """Unnormalizes the input."""
        b, c, h, w = x.shape
        input_data = x.reshape(b, self.norm_groups, -1)
        return (input_data * std + mean).reshape(b, c, h, w)

    def forward(
        self,
        x_stack: List[torch.Tensor],
        iscomplex: bool = False,
        pad_sizes: Tuple[List[int], List[int], int, int] = None,
        mean: torch.Tensor = None,
        std: torch.Tensor = None,
    ) -> torch.Tensor:
        """Forward pass of :class:`UnetDecoder`.

        Parameters
        ----------
        x_stack : List[torch.Tensor]
            List of tensors from the encoder.
        iscomplex : bool, optional
            Whether the input is complex. Default is ``False``.
        pad_sizes : Tuple[List[int], List[int], int, int], optional
            Padding sizes. Default is ``None``.
        mean : torch.Tensor, optional
            Mean of the input. Default is ``None``.
        std : torch.Tensor, optional
            Standard deviation of the input. Default is ``None``.

        Returns
        -------
        torch.Tensor
            Output of the network.
        """
        output = x_stack.pop()
        # apply up-sampling layers
        for transpose_conv, conv in zip(self.up_transpose_conv, self.up_conv):
            downsample_layer = x_stack.pop()
            output = transpose_conv(output)

            # reflect pad on the right/bottom if needed to handle odd input dimensions
            padding = [0, 0, 0, 0]
            if output.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # padding right
            if output.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # padding bottom
            if torch.sum(torch.tensor(padding)) != 0:
                output = torch.nn.functional.pad(output, padding, "reflect")

            output = torch.cat([output, downsample_layer], dim=1)
            output = conv(output)

        if self.padding:
            output = self.unpad(output, *pad_sizes)  # type: ignore
        if self.normalize and mean is not None and std is not None:
            output = self.unnorm(output, mean, std)
        if iscomplex:
            output = self.chan_complex_to_last_dim(output)

        return output



# coding=utf-8
# ! /usr/bin/python
import importlib.util
import subprocess
import sys
from pathlib import Path

import setuptools
from setuptools import Command
from setuptools import logging as setuptools_logging

__author__ = "Dimitris Karkalousos"


spec = importlib.util.spec_from_file_location("package_info", "atommic/package_info.py")
package_info = importlib.util.module_from_spec(spec)  # type: ignore
spec.loader.exec_module(package_info)  # type: ignore

__contact_emails__ = package_info.__contact_emails__
__contact_names__ = package_info.__contact_names__
__description__ = package_info.__description__
__download_url__ = package_info.__download_url__
__homepage__ = package_info.__homepage__
__keywords__ = package_info.__keywords__
__license__ = package_info.__license__
__package_name__ = package_info.__package_name__
__repository_url__ = package_info.__repository_url__
__version__ = package_info.__version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
long_description_content_type = "text/markdown"


###############################################################################
#                             Dependency Loading                              #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #


def _load_requirements(requirements_file, folder="requirements"):
    """Load requirements from a file."""
    requirements = []
    with open(Path(folder) / Path(requirements_file), "r") as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#"):
                requirements.append(line)
    return requirements


install_requires = _load_requirements("requirements.txt")


###############################################################################
#                            Code style checkers                              #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #


class StyleCommand(Command):
    """Run code style checkers."""

    __LINE_WIDTH = 119
    __ISORT_BASE = (
        "isort "
        # These two lines makes isort compatible with black.
        "--multi-line=3 --trailing-comma --force-grid-wrap=0 "
        f"--use-parentheses --line-width={__LINE_WIDTH} -rc -ws"
    )
    __BLACK_BASE = f"black --skip-string-normalization --line-length={__LINE_WIDTH}"
    description = "Run code style checkers."
    user_options = [
        ("scope=", None, "Folder of file to operate within."),
        ("fix", None, "True if tries to fix issues in-place."),
    ]

    def __call_checker(self, base_command, scope, check):
        command = list(base_command)

        command.append(scope)

        if check:
            command.extend(["--check", "--diff"])

        self.announce(msg=f'Running command: {" ".join(command)}', level=setuptools_logging.INFO)

        return subprocess.call(command)

    def _isort(self, scope, check):
        return self.__call_checker(base_command=self.__ISORT_BASE.split(), scope=scope, check=check)

    def _black(self, scope, check):
        return self.__call_checker(base_command=self.__BLACK_BASE.split(), scope=scope, check=check)

    def _pass(self):
        self.announce(msg="\033[32mPASS\x1b[0m", level=setuptools_logging.INFO)

    def _fail(self):
        self.announce(msg="\033[31mFAIL\x1b[0m", level=setuptools_logging.INFO)

    # noinspection PyAttributeOutsideInit
    def initialize_options(self):
        self.scope = "."
        self.fix = ""

    def run(self):
        scope, check = self.scope, not self.fix
        isort_return = self._isort(scope=scope, check=check)
        black_return = self._black(scope=scope, check=check)

        if isort_return == 0 and black_return == 0:
            self._pass()
        else:
            self._fail()
            sys.exit(isort_return if isort_return != 0 else black_return)

    def finalize_options(self):
        raise NotImplementedError()


###############################################################################
#                             Setup                                           #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #


setuptools.setup(
    name=__package_name__,
    # Versions should comply with PEP440.  For a discussion on single-sourcing the version across setup.py and the
    # project code, see https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,
    description=__description__,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    # The project's main homepage.
    url=__repository_url__,
    download_url=__download_url__,
    # Author details
    author=__contact_names__,
    author_email=__contact_emails__,
    # maintainer Details
    maintainer=__contact_names__,
    maintainer_email=__contact_emails__,
    # The licence under which the project is released
    license=__license__,
    classifiers=[
        # How mature is this project? Common values are
        #  1 - Planning
        #  2 - Pre-Alpha
        #  3 - Alpha
        #  4 - Beta
        #  5 - Production/Stable
        #  6 - Mature
        #  7 - Inactive
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        # Indicate what your project relates to
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Supported python versions
        "Programming Language :: Python :: 3.10",
        # Additional Setting
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include=["atommic", "atommic.*"]),
    install_requires=install_requires,
    setup_requires=["pytest-runner"],
    python_requires=">=3.10, <=3.12",
    # Add in any packaged data.
    include_package_data=True,
    exclude=["tools", "tests"],
    package_data={"": ["*.tsv", "*.txt", "*.far", "*.fst", "*.cpp", "Makefile"]},
    zip_safe=False,
    # PyPI package information.
    keywords=__keywords__,
    # Custom commands.
    cmdclass={"style": StyleCommand},
    # entry points
    entry_points={
        "console_scripts": [
            "atommic = atommic.cli:main",
        ],
    },
)

###############################################################################
#                             End of File                                    #
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% #



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import Conv2d
from atommic.collections.reconstruction.nn.mwcnn_base.mwcnn_block import MWCNN
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.collections.reconstruction.nn.vsnet_base.vsnet_block import (
    DataConsistencyLayer,
    VSNetBlock,
    WeightedAverageTerm,
)
from atommic.core.classes.common import typecheck

__all__ = ["VSNet"]


class VSNet(BaseMRIReconstructionModel):
    """Implementation of the Variable-Splitting Net, as presented in [Duan2019]_.

    References
    ----------
    .. [Duan2019] Duan, J. et al. (2019) Vs-net: Variable splitting network for accelerated parallel MRI
        reconstruction, Lecture Notes in Computer Science (including subseries Lecture Notes in Artificial
        Intelligence and Lecture Notes in Bioinformatics), 11767 LNCS, pp. 713–722. doi: 10.1007/978-3-030-32251-9_78.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`VSNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        num_cascades = cfg_dict.get("num_cascades")
        self.num_cascades = cfg_dict.get("num_cascades")

        image_model_architecture = cfg_dict.get("imspace_model_architecture")
        if image_model_architecture == "CONV":
            image_model = Conv2d(
                in_channels=cfg_dict.get("imspace_in_channels", 2),
                out_channels=cfg_dict.get("imspace_out_channels", 2),
                hidden_channels=cfg_dict.get("imspace_conv_hidden_channels"),
                n_convs=cfg_dict.get("imspace_conv_n_convs"),
                batchnorm=cfg_dict.get("imspace_conv_batchnorm"),
            )
        elif image_model_architecture == "MWCNN":
            image_model = MWCNN(
                input_channels=cfg_dict.get("imspace_in_channels", 2),
                first_conv_hidden_channels=cfg_dict.get("image_mwcnn_hidden_channels"),
                num_scales=cfg_dict.get("image_mwcnn_num_scales"),
                bias=cfg_dict.get("image_mwcnn_bias"),
                batchnorm=cfg_dict.get("image_mwcnn_batchnorm"),
            )
        elif image_model_architecture in ["UNET", "NORMUNET"]:
            image_model = NormUnet(
                cfg_dict.get("imspace_unet_num_filters"),
                cfg_dict.get("imspace_unet_num_pool_layers"),
                in_chans=cfg_dict.get("imspace_in_channels", 2),
                out_chans=cfg_dict.get("imspace_out_channels", 2),
                drop_prob=cfg_dict.get("imspace_unet_dropout_probability"),
                padding_size=cfg_dict.get("imspace_unet_padding_size"),
                normalize=cfg_dict.get("imspace_unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "VSNet is currently implemented only with image_model_architecture == 'MWCNN' or 'UNet'."
                f"Got {image_model_architecture}."
            )

        image_model = torch.nn.ModuleList([image_model] * num_cascades)
        data_consistency_model = torch.nn.ModuleList([DataConsistencyLayer()] * num_cascades)
        weighted_average_model = torch.nn.ModuleList([WeightedAverageTerm()] * num_cascades)

        self.reconstruction_module = VSNetBlock(
            denoiser_block=image_model,
            data_consistency_block=data_consistency_model,
            weighted_average_block=weighted_average_model,
            num_cascades=num_cascades,
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim,
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`VSNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        return check_stacked_complex(
            coil_combination_method(
                ifft2(
                    self.reconstruction_module(y, sensitivity_maps, mask),
                    self.fft_centered,
                    self.fft_normalization,
                    self.spatial_dims,
                ),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet
from atommic.collections.segmentation.nn.vnet_base.vnet_block import VNet
from atommic.core.classes.common import typecheck

__all__ = ["SegmentationVNet"]


class SegmentationVNet(BaseSegmentationNet):
    """Implementation of the V-Net for MRI segmentation, as presented in [Milletari2016]_.

    References
    ----------
    .. [Milletari2016] Fausto Milletari, Nassir Navab, Seyed-Ahmad Ahmadi. V-Net: Fully Convolutional Neural Networks
        for Volumetric Medical Image Segmentation, 2016. https://arxiv.org/abs/1606.04797

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        return VNet(
            in_chans=self.input_channels,
            out_chans=cfg.get("segmentation_module_output_channels", 2),
            act=cfg.get("segmentation_module_activation", "elu"),
            drop_prob=cfg.get("segmentation_module_dropout", 0.0),
            bias=cfg.get("segmentation_module_bias", False),
        )

    @typecheck()
    def forward(self, image: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of :class:`BaseSegmentationNet`.

        Parameters
        ----------
        image : torch.Tensor
            Input image. Shape [batch_size, n_x, n_y] or [batch_size, n_x, n_y, 2]
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Predicted segmentation. Shape [batch_size, n_classes, n_x, n_y]
        """
        if self.consecutive_slices > 1:
            batch, slices = image.shape[:2]
            image = image.reshape(batch * slices, *image.shape[2:])

        if image.shape[-1] == 2:
            if self.input_channels == 1:
                image = torch.view_as_complex(image).unsqueeze(1)
                if self.magnitude_input:
                    image = torch.abs(image)
            elif self.input_channels == 2 and not self.magnitude_input:
                image = image.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        elif self.magnitude_input:
            image = torch.abs(image)

        if image.dim() == 3:
            image = image.unsqueeze(1)

        # if dim 1 is even, add a row of zeros to make it odd
        if image.shape[1] % 2 != 0 and image.shape[1] != 1:
            image = torch.cat((image, torch.zeros_like(image[:, 0:1, :, :]).to(image.device)), dim=1)

        mean = 1.0
        std = 1.0
        if self.normalize:
            image, mean, std = self.norm(image)
        image, pad_sizes = self.pad(image)
        segmentation = self.segmentation_module(image)
        segmentation = self.unpad(segmentation, *pad_sizes)
        if self.normalize:
            segmentation = self.unnorm(segmentation, mean, std)

        if self.normalize_segmentation_output:
            segmentation = (segmentation - segmentation.min()) / (segmentation.max() - segmentation.min())

        if self.consecutive_slices > 1:
            segmentation = segmentation.reshape(batch, slices, *segmentation.shape[1:])

        return torch.abs(segmentation)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/Project-MONAI/MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/transformerblock.py#L18

from typing import Tuple, Union

import torch
from einops.layers.torch import Rearrange
from torch import nn

SUPPORTED_DROPOUT_MODE = {"vit", "swin"}


class MLPBlock(nn.Module):
    """Implementation of a multi-layer perceptron block, as presented in [Dosovitskiy2020]_.

    References
    ----------
    .. [Dosovitskiy2020] Dosovitskiy A, Beyer L, Kolesnikov A, Weissenborn D, Zhai X, Unterthiner T, Dehghani M,
        Minderer M, Heigold G, Gelly S, Uszkoreit J. An image is worth 16x16 words: Transformers for image recognition
        at scale. arXiv preprint arXiv:2010.11929. 2020 Oct 22.

    .. note::
        This is a wrapper for monai implementation of a multi-layer perceptron block. See:
        https://github.com/Project-MONAI/MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/
        transformerblock.py#L18
    """

    def __init__(
        self,
        hidden_size: int,
        mlp_dim: int,
        dropout_rate: float = 0.0,
        act: Union[Tuple, str] = "GELU",  # pylint: disable=unused-argument
        dropout_mode="vit",
    ):
        """Inits :class:`MLPBlock`.

        Parameters
        ----------
        hidden_size : int
            Dimension of hidden layer.
        mlp_dim : int
            Dimension of MLP layer.
        dropout_rate : float, optional
            Faction of the input units to drop. Default is ``0.0``.
        act : Union[Tuple, str], optional
            Activation type and arguments. Default is ``"GELU"``.
        dropout_mode : str, optional
            Dropout mode, can be "vit" or "swin". Default is ``vit``.
            - "vit" mode uses two dropout instances as implemented in
            https://github.com/google-research/vision_transformer/blob/main/vit_jax/models.py#L87
            - "swin" corresponds to one instance as implemented in
            https://github.com/microsoft/Swin-Transformer/blob/main/models/swin_mlp.py#L23
        """
        super().__init__()
        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")
        mlp_dim = mlp_dim or hidden_size
        self.linear1 = nn.Linear(hidden_size, mlp_dim)
        self.linear2 = nn.Linear(mlp_dim, hidden_size)
        self.fn = nn.GELU()
        self.drop1 = nn.Dropout(dropout_rate)
        if dropout_mode == "vit":
            self.drop2 = nn.Dropout(dropout_rate)
        elif dropout_mode == "swin":
            self.drop2 = self.drop1
        else:
            raise ValueError(f"dropout_mode should be one of {SUPPORTED_DROPOUT_MODE}")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`MLPBlock`."""
        return self.drop2(self.linear2(self.drop1(self.fn(self.linear1(x)))))


class SABlock(nn.Module):
    """Implementation of a self-attention block, as presented in [Dosovitskiy2020]_.

    References
    ----------
    .. [Dosovitskiy2020] Dosovitskiy A, Beyer L, Kolesnikov A, Weissenborn D, Zhai X, Unterthiner T, Dehghani M,
        Minderer M, Heigold G, Gelly S, Uszkoreit J. An image is worth 16x16 words: Transformers for image recognition
        at scale. arXiv preprint arXiv:2010.11929. 2020 Oct 22.

    .. note::
        This is a wrapper for monai implementation of self-attention block. See: https://github.com/Project-MONAI/
        MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/transformerblock.py#L18
    """

    def __init__(self, hidden_size: int, num_heads: int, dropout_rate: float = 0.0, qkv_bias: bool = False):
        """Inits :class:`SABlock`.

        Parameters
        ----------
        hidden_size : int
            Dimension of hidden layer.
        num_heads : int
            Number of attention heads.
        dropout_rate : float, optional
            Faction of the input units to drop. Default is ``0.0``.
        qkv_bias : bool, optional
            Bias term for the qkv linear layer. Default is ``False``.
        """
        super().__init__()
        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")
        if hidden_size % num_heads != 0:
            raise ValueError("hidden size should be divisible by num_heads.")
        self.num_heads = num_heads
        self.out_proj = nn.Linear(hidden_size, hidden_size)
        self.qkv = nn.Linear(hidden_size, hidden_size * 3, bias=qkv_bias)
        self.input_rearrange = Rearrange("b h (qkv l d) -> qkv b l h d", qkv=3, l=num_heads)
        self.out_rearrange = Rearrange("b h l d -> b l (h d)")
        self.drop_output = nn.Dropout(dropout_rate)
        self.drop_weights = nn.Dropout(dropout_rate)
        self.head_dim = hidden_size // num_heads
        self.scale = self.head_dim**-0.5

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`SABlock`."""
        output = self.input_rearrange(self.qkv(x))
        q, k, v = output[0], output[1], output[2]
        att_mat = (torch.einsum("blxd,blyd->blxy", q, k) * self.scale).softmax(dim=-1)
        att_mat = self.drop_weights(att_mat)
        x = torch.einsum("bhxy,bhyd->bhxd", att_mat, v)
        x = self.out_rearrange(x)
        x = self.out_proj(x)
        x = self.drop_output(x)
        return x


class TransformerBlock(nn.Module):
    """Implementation of a transformer block, as presented in [Dosovitskiy2020]_.

    References
    ----------
    .. [Dosovitskiy2020] Dosovitskiy A, Beyer L, Kolesnikov A, Weissenborn D, Zhai X, Unterthiner T, Dehghani M,
        Minderer M, Heigold G, Gelly S, Uszkoreit J. An image is worth 16x16 words: Transformers for image recognition
        at scale. arXiv preprint arXiv:2010.11929. 2020 Oct 22.

    .. note::
        This is a wrapper for monai implementation of self-attention block. See: https://github.com/Project-MONAI/
        MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/transformerblock.py#L18
    """

    def __init__(
        self,
        hidden_size: int,
        mlp_dim: int,
        num_heads: int,
        dropout_rate: float = 0.0,
        qkv_bias: bool = False,
        spatial_dims: int = 2,
    ):
        """Inits :class:`TransformerBlock`.

        Parameters
        ----------
        hidden_size : int
            Dimension of hidden layer.
        mlp_dim : int
            Dimension of the mlp layer.
        num_heads : int
            Number of attention heads.
        dropout_rate : float, optional
            Faction of the input units to drop. Default is ``0.0``.
        qkv_bias : bool, optional
            Bias term for the qkv linear layer. Default is ``False``.
        spatial_dims : int, optional
            Number of spatial dimensions. Default is ``2``.
        """
        super().__init__()
        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")
        if hidden_size % num_heads != 0:
            raise ValueError("hidden_size should be divisible by num_heads.")

        self.mlp = MLPBlock(hidden_size, mlp_dim, dropout_rate)
        self.attn = SABlock(hidden_size, num_heads, dropout_rate, qkv_bias)
        if spatial_dims == 2:
            self.norm1 = nn.InstanceNorm2d(hidden_size)
            self.norm2 = nn.InstanceNorm2d(hidden_size)
        elif spatial_dims == 3:
            self.norm1 = nn.InstanceNorm3d(hidden_size)
            self.norm2 = nn.InstanceNorm3d(hidden_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`TransformerBlock`."""
        x = x + self.attn(self.norm1(x))
        x = x + self.mlp(self.norm2(x))
        return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import warnings

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import Conv2d
from atommic.collections.reconstruction.nn.crossdomain_base.crossdomain_block import CrossDomainNetwork, MultiCoil
from atommic.collections.reconstruction.nn.didn_base.didn_block import DIDN
from atommic.collections.reconstruction.nn.mwcnn_base.mwcnn_block import MWCNN
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["XPDNet"]


class XPDNet(BaseMRIReconstructionModel):
    """Implementation of the XPDNet, as presented in [Ramzi2021]_.

    References
    ----------
    .. [Ramzi2021] Ramzi, Zaccharie, et al. “XPDNet for MRI Reconstruction: An Application to the 2020 FastMRI
        Challenge. ArXiv:2010.07290 [Physics, Stat], July 2021. arXiv.org, http://arxiv.org/abs/2010.07290.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`XPDNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        num_primal = cfg_dict.get("num_primal")
        num_dual = cfg_dict.get("num_dual")
        num_iter = cfg_dict.get("num_iter")

        kspace_model_architecture = cfg_dict.get("kspace_model_architecture")
        dual_conv_hidden_channels = cfg_dict.get("dual_conv_hidden_channels", 64)
        dual_conv_num_dubs = cfg_dict.get("dual_conv_num_dubs", 2)
        dual_conv_batchnorm = cfg_dict.get("dual_conv_batchnorm", True)
        dual_didn_hidden_channels = cfg_dict.get("dual_didn_hidden_channels", 64)
        dual_didn_num_dubs = cfg_dict.get("dual_didn_num_dubs", 2)
        dual_didn_num_convs_recon = cfg_dict.get("dual_didn_num_convs_recon", True)

        if cfg_dict.get("use_primal_only"):
            kspace_model_list = None
            num_dual = 1
        elif kspace_model_architecture == "CONV":
            kspace_model_list = torch.nn.ModuleList(
                [
                    MultiCoil(
                        Conv2d(
                            cfg_dict.get("kspace_in_channels") * (num_dual + num_primal + 1),
                            cfg_dict.get("kspace_out_channels") * num_dual,
                            dual_conv_hidden_channels,
                            dual_conv_num_dubs,
                            batchnorm=dual_conv_batchnorm,
                        )
                    )
                    for _ in range(num_iter)
                ]
            )
        elif kspace_model_architecture == "DIDN":
            kspace_model_list = torch.nn.ModuleList(
                [
                    MultiCoil(
                        DIDN(
                            in_channels=cfg_dict.get("kspace_in_channels") * (num_dual + num_primal + 1),
                            out_channels=cfg_dict.get("kspace_out_channels") * num_dual,
                            hidden_channels=dual_didn_hidden_channels,
                            num_dubs=dual_didn_num_dubs,
                            num_convs_recon=dual_didn_num_convs_recon,
                        )
                    )
                    for _ in range(num_iter)
                ]
            )
        elif kspace_model_architecture in ["UNET", "NORMUNET"]:
            kspace_model_list = torch.nn.ModuleList(
                [
                    MultiCoil(
                        NormUnet(
                            cfg_dict.get("kspace_unet_num_filters"),
                            cfg_dict.get("kspace_unet_num_pool_layers"),
                            in_chans=cfg_dict.get("kspace_in_channels") * (num_dual + num_primal + 1),
                            out_chans=cfg_dict.get("kspace_out_channels") * num_dual,
                            drop_prob=cfg_dict.get("kspace_unet_dropout_probability"),
                            padding_size=cfg_dict.get("kspace_unet_padding_size"),
                            normalize=cfg_dict.get("kspace_unet_normalize"),
                        ),
                        coil_to_batch=True,
                    )
                    for _ in range(num_iter)
                ]
            )
        else:
            raise NotImplementedError(
                "XPDNet is currently implemented for kspace_model_architecture == 'CONV' or 'DIDN'."
                f"Got kspace_model_architecture == {kspace_model_architecture}."
            )

        image_model_architecture = cfg_dict.get("image_model_architecture")
        mwcnn_hidden_channels = cfg_dict.get("mwcnn_hidden_channels", 16)
        mwcnn_num_scales = cfg_dict.get("mwcnn_num_scales", 2)
        mwcnn_bias = cfg_dict.get("mwcnn_bias", True)
        mwcnn_batchnorm = cfg_dict.get("mwcnn_batchnorm", True)

        if image_model_architecture == "MWCNN":
            image_model_list = torch.nn.ModuleList(
                [
                    torch.nn.Sequential(
                        MWCNN(
                            input_channels=cfg_dict.get("imspace_in_channels") * (num_primal + num_dual),
                            first_conv_hidden_channels=mwcnn_hidden_channels,
                            num_scales=mwcnn_num_scales,
                            bias=mwcnn_bias,
                            batchnorm=mwcnn_batchnorm,
                        ),
                        torch.nn.Conv2d(2 * (num_primal + num_dual), 2 * num_primal, kernel_size=3, padding=1),
                    )
                    for _ in range(num_iter)
                ]
            )
        elif image_model_architecture in ["UNET", "NORMUNET"]:
            image_model_list = torch.nn.ModuleList(
                [
                    NormUnet(
                        cfg_dict.get("imspace_unet_num_filters"),
                        cfg_dict.get("imspace_unet_num_pool_layers"),
                        in_chans=cfg_dict.get("imspace_in_channels") * (num_primal + num_dual),
                        out_chans=cfg_dict.get("imspace_out_channels") * num_primal,
                        drop_prob=cfg_dict.get("imspace_unet_dropout_probability"),
                        padding_size=cfg_dict.get("imspace_unet_padding_size"),
                        normalize=cfg_dict.get("imspace_unet_normalize"),
                    )
                    for _ in range(num_iter)
                ]
            )
        else:
            raise NotImplementedError(f"Image model architecture {image_model_architecture} not found for XPDNet.")

        self.num_cascades = cfg_dict.get("num_cascades")

        self.reconstruction_module = CrossDomainNetwork(
            image_model_list=image_model_list,
            kspace_model_list=kspace_model_list,
            domain_sequence="KI" * num_iter,
            image_buffer_size=num_primal,
            kspace_buffer_size=num_dual,
            normalize_image=cfg_dict.get("normalize_image"),
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim,
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`XPDNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        prediction = self.reconstruction_module(y, sensitivity_maps, mask)
        # filter UserWarning: ComplexHalf support is experimental and many operators don't support it yet.
        # TODO: remove this when PyTorch fixes the issue.
        warnings.filterwarnings("ignore", category=UserWarning)
        return torch.view_as_real(prediction[..., 0] + 1j * prediction[..., 1])



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/callbacks/preemption.py

import signal
import sys
import warnings

import torch
from pytorch_lightning.callbacks import Callback


class PreemptionCallback(Callback):
    """PreemptionCallback class creates a callback that checks for preemption during training at the end of every step.
    Upon preemption the callback provides a function to gracefully exit the training immediately and also saves the
    current state in a checkpoint as *last.ckpt. (to be able to start from the same step without wasting any compute
    while resuming the next time).

    PreemptionCallback is always enabled by default via the arg create_preemption_callback under ExpManagerConfig.
    To disable please pass create_preemption_callback: False in your config file.
    """

    def __init__(self, checkpoint_callback, sig=None):
        """Inits :class:`PreemptionCallback`.

        Parameters
        ----------
        checkpoint_callback : pytorch_lightning.callbacks.ModelCheckpoint
            The checkpoint callback
        sig : int, optional
            The signal to be used for preemption, by default None
        """
        self.sig = sig
        if self.sig is None:
            self.sig = signal.SIGTERM
        self.checkpoint_callback = checkpoint_callback
        self.preemption_enabled = False

    @property
    def interrupted(self):
        """Checks if the job was preempted by broadcasting the preemption signal to all ranks."""
        interrupted = torch.tensor(self._interrupted, device=torch.cuda.current_device(), dtype=torch.int32)
        torch.distributed.broadcast(interrupted, 0)
        interrupted = bool(interrupted.item())
        return interrupted

    def on_train_start(self, trainer, pl_module):
        """
        Defines custom handlers at the beginning of training to be executed when the preemption signal is received.
        """
        # Check if torch distributed is initialised, as It's needed for broadcasting the preemption signal to all ranks
        if not (torch.distributed.is_available() and torch.distributed.is_initialized()):
            warnings.warn("Preemption requires torch distributed to be initialized, disabling preemption callback")
        else:
            self.preemption_enabled = True
            # Bool var that's initialized to false and made True upon receving the preemption signal
            self._interrupted = False
            self.released = False
            self.original_handler = signal.getsignal(self.sig)

            # Master handler executed only by rank 0 when the preemption siganal is received,
            # to avoid deadlock conditions
            def master_handler(signum, frame):  # pylint: disable=unused-argument
                """Handler executed by rank 0 when the preemption signal is received."""
                self.release()
                self._interrupted = True

            # Handler executed by the non zero ranks
            def ignoring_handler(signum, frame):  # pylint: disable=unused-argument
                """Handler executed by non zero ranks when the preemption signal is received."""
                self.release()

            self.private_rank = torch.distributed.get_rank()
            if self.private_rank == 0:
                signal.signal(self.sig, master_handler)
            else:
                signal.signal(self.sig, ignoring_handler)

        return self

    def on_train_end(self, trainer, pl_module):
        """Defines custom handlers at the end of training to be executed when the preemption signal is received.

        Parameters
        ----------
        trainer : pytorch_lightning.Trainer
            The trainer object
        pl_module : pytorch_lightning.LightningModule
            The lightning module
        """
        if self.preemption_enabled:
            self.release()

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx: int):
        """Defines custom handlers at the end of every training step to be executed when the preemption signal is
        received.

        Parameters
        ----------
        trainer : pytorch_lightning.Trainer
            The trainer object
        pl_module : pytorch_lightning.LightningModule
            The lightning module
        outputs : list
            The outputs of the training step
        batch : list
            The batch of data
        batch_idx : int
            The index of the batch
        """
        if self.preemption_enabled:
            # check if the job was preempted at the end of every training step/iteration
            # NOTE: "self.interrupted" is a property which triggers a distributed broadcast of "_interrupted" flag from
            # rank 0 to all other ranks, to avoid performance overheads it's best to store the result in a regular
            # local variable
            interrupted = self.interrupted
            if interrupted:
                warnings.warn("Received SIGTERM, saving checkpoint and exiting")
                monitor_candidates = self.checkpoint_callback._monitor_candidates(  # pylint: disable=protected-access
                    trainer
                )
                self.checkpoint_callback._save_last_checkpoint(  # pylint: disable=protected-access
                    trainer, monitor_candidates
                )
                sys.exit(0)

    def release(self):
        """Releases the preemption callback."""
        if self.released:
            return False

        signal.signal(self.sig, self.original_handler)
        self.released = True
        return True



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/config/pytorch_lightning.py

from dataclasses import dataclass
from typing import Any, Optional

from hydra.core.config_store import ConfigStore

__all__ = ["TrainerConfig"]

cs = ConfigStore.instance()


@dataclass
class TrainerConfig:
    """TrainerConfig is a dataclass that holds all the hyperparameters for the training process."""

    logger: Any = True
    callbacks: Optional[Any] = None
    default_root_dir: Optional[str] = None
    gradient_clip_val: float = 0
    num_nodes: int = 1
    enable_progress_bar: bool = True
    overfit_batches: Any = 0.0
    check_val_every_n_epoch: int = 1
    fast_dev_run: bool = False
    accumulate_grad_batches: Any = 1
    max_epochs: int = 1000
    min_epochs: int = 1
    max_steps: Optional[int] = -1
    min_steps: Optional[int] = None
    limit_train_batches: Any = 1.0
    limit_val_batches: Any = 1.0
    limit_test_batches: Any = 1.0
    val_check_interval: Any = 1.0
    log_every_n_steps: int = 50
    accelerator: Optional[str] = None
    sync_batchnorm: bool = False
    precision: Any = 32
    num_sanity_val_steps: int = 2
    profiler: Optional[Any] = None
    benchmark: bool = False
    deterministic: bool = False
    use_distributed_sampler: bool = True
    detect_anomaly: bool = False
    plugins: Optional[Any] = None  # Optional[Union[str, list]]
    limit_predict_batches: float = 1.0
    gradient_clip_algorithm: str = 'norm'
    max_time: Optional[Any] = None  # can be one of Union[str, timedelta, Dict[str, int], None]
    reload_dataloaders_every_n_epochs: int = 0
    devices: Any = None
    strategy: Any = None
    enable_checkpointing: bool = False
    enable_model_summary: bool = True
    inference_mode: bool = True
    barebones: bool = False


# Register the trainer config.
cs.store(group="trainer", name="trainer", node=TrainerConfig)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from omegaconf import DictConfig

from atommic.collections.segmentation.nn.segmentationnet import BaseSegmentationNet
from atommic.collections.segmentation.nn.unetr_base.unetr_block import UNETR

__all__ = ["SegmentationUNetR"]

from atommic.core.classes.common import typecheck


class SegmentationUNetR(BaseSegmentationNet):
    """Implementation of the UNETR for MRI segmentation, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    """

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.

        Returns
        -------
        torch.nn.Module
            The segmentation module.
        """
        return UNETR(
            in_channels=self.input_channels,
            out_channels=cfg.get("segmentation_module_output_channels", 2),
            img_size=cfg.get("segmentation_module_img_size", (256, 256)),
            feature_size=cfg.get("segmentation_module_channels", 64),
            hidden_size=cfg.get("segmentation_module_hidden_size", 768),
            mlp_dim=cfg.get("segmentation_module_mlp_dim", 3072),
            num_heads=cfg.get("segmentation_module_num_heads", 12),
            pos_embed=cfg.get("segmentation_module_pos_embed", "conv"),
            norm_name=cfg.get("segmentation_module_norm_name", "instance"),
            conv_block=cfg.get("segmentation_module_conv_block", True),
            res_block=cfg.get("segmentation_module_res_block", True),
            dropout_rate=cfg.get("segmentation_module_dropout", 0.0),
            spatial_dims=cfg.get("dimensionality", 2),
            qkv_bias=cfg.get("segmentation_module_qkv_bias", False),
        )

    @typecheck()
    def forward(self, image: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Forward pass of the network.

        Parameters
        ----------
        image : torch.Tensor
            Input image. Shape [batch_size, slices, classes, n_x, n_y] or [batch_size, slices, classes, n_x, n_y, 2]
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Predicted segmentation. Shape [batch_size, n_classes, n_x, n_y]
        """
        # Adjust the dimensions of the input image
        if image.shape[-1] == 2:
            if self.input_channels == 1:
                image = torch.view_as_complex(image).unsqueeze(1)
                if self.magnitude_input:
                    image = torch.abs(image)
            elif self.input_channels == 2 and not self.magnitude_input:
                image = image.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")

        if image.dim() == 4:
            # we are missing the classes dimension
            image = image.unsqueeze(2)

        mean = 1.0
        std = 1.0
        if self.normalize:
            image, mean, std = self.norm(image)
        image, pad_sizes = self.pad(image)
        if self.consecutive_slices > 1:
            batch, slices = image.shape[:2]
            image = image.reshape(batch * slices, *image.shape[2:])
        segmentation = self.segmentation_module(image)
        segmentation = self.unpad(segmentation, *pad_sizes)
        if self.normalize:
            segmentation = self.unnorm(segmentation, mean, std)

        if self.normalize_segmentation_output:
            segmentation = (segmentation - segmentation.min()) / (segmentation.max() - segmentation.min())

        if self.consecutive_slices > 1:
            segmentation = segmentation.reshape(batch, slices, *segmentation.shape[1:])

        return torch.abs(segmentation)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/formatters/utils.py

import sys

from atommic.constants import ATOMMIC_ENV_VARNAME_ENABLE_COLORING
from atommic.utils.env_var_parsing import get_envbool

__all__ = ["check_color_support", "to_unicode"]


def check_color_support():
    """Checks if the terminal supports color.

    Returns
    -------
    bool
        True if the terminal supports color, False otherwise.
    """
    # Colors can be forced with an env variable
    return bool(not sys.platform.lower().startswith("win") and get_envbool(ATOMMIC_ENV_VARNAME_ENABLE_COLORING, False))


def to_unicode(value):
    """Converts a string to unicode. If the string is already unicode, it is returned as is. If it is a byte string, it
     is decoded using utf-8.

    Parameters
    ----------
    value : str
        The string to convert.

    Returns
    -------
    str
        The converted string.
    """
    try:
        if isinstance(value, (str, type(None))):
            return value

        if not isinstance(value, bytes):
            raise TypeError(f"Expected bytes, unicode, or None; got %{type(value)}")

        return value.decode("utf-8")

    except UnicodeDecodeError:
        return repr(value)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.losses.cross_entropy import CrossEntropyLoss  # noqa: F401
from atommic.collections.segmentation.losses.dice import Dice  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/callbacks/callbacks.py

import time

from pytorch_lightning.callbacks import Callback
from pytorch_lightning.utilities import rank_zero_only


class LogEpochTimeCallback(Callback):
    """Simple callback that logs how long each epoch takes, in seconds, to a pytorch lightning log.

    .. note::
        Extends :class:`pytorch_lightning.callbacks.Callback`.

    Examples
    --------
    >>> from pytorch_lightning import Trainer
    >>> from pytorch_lightning.callbacks import ModelCheckpoint
    >>> from pytorch_lightning.loggers import TensorBoardLogger
    >>> from atommic.collections.common.callbacks.callbacks import LogEpochTimeCallback
    >>> logger = TensorBoardLogger("tb_logs", name="my_model")
    >>> checkpoint_callback = ModelCheckpoint(
    ...     dirpath="checkpoints",
    ...     filename="my_model-{epoch:02d}-{val_loss:.2f}",
    ...     save_top_k=3,
    ...     verbose=True,
    ...     monitor="val_loss",
    ...     mode="min",
    ... )
    >>> trainer = Trainer(
    ...     logger=logger,
    ...     callbacks=[LogEpochTimeCallback(), checkpoint_callback],
    ...     max_epochs=10,
    ...     gpus=1,
    ...     strategy="ddp_fork",
    ...     accelerator="gpu",
    ...     precision=16,
    ... )
    """

    def __init__(self):
        """Inits :class:`LogEpochTimeCallback`."""
        super().__init__()
        self.epoch_start = time.time()

    @rank_zero_only
    def on_train_epoch_start(self, trainer, pl_module):
        """Called at the start of each epoch.

        Parameters
        ----------
        trainer : pytorch_lightning.Trainer
            The trainer object.
        pl_module : pytorch_lightning.LightningModule
            The lightning module.
        """
        self.epoch_start = time.time()

    @rank_zero_only
    def on_train_epoch_end(self, trainer, pl_module):
        """Called at the end of each epoch.

        Parameters
        ----------
        trainer : pytorch_lightning.Trainer
            The trainer object.
        pl_module : pytorch_lightning.LightningModule
            The lightning module.
        """
        curr_time = time.time()
        duration = curr_time - self.epoch_start
        trainer.logger.log_metrics({"epoch_time": duration}, step=trainer.global_step)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/classes/dataset.py
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from torch.utils import data

from atommic.core.classes.common import Serialization, Typing, typecheck

__all__ = ["Dataset", "DatasetConfig", "IterableDataset"]


class Dataset(data.Dataset, Typing, Serialization, ABC):
    """Dataset with output ports. Please Note: Subclasses of IterableDataset should *not* implement input_types."""

    @staticmethod
    def _collate_fn(batch):
        """A default implementation of a collation function. Users should override this method to define custom data
        loaders.
        """
        return data.dataloader.default_collate(batch)

    @typecheck()
    def collate_fn(self, batch):
        """This is the method that user pass as functor to DataLoader.
        The method optionally performs neural type checking and add types to the outputs.

        Please note, subclasses of Dataset should not implement `input_types`.

        # Usage:

        .. code-block::

            dataloader = torch.utils.data.DataLoader(
                    ....,
                    collate_fn=dataset.collate_fn,
                    ....
            )

        Returns
        -------
        Collated batch, with or without types.
        """
        if self.input_types is not None:
            raise TypeError("Datasets should not implement `input_types` as they are not checked")

        # Simply forward the inner `_collate_fn`
        return self._collate_fn(batch)


class IterableDataset(data.IterableDataset, Typing, Serialization, ABC):
    """Iterable Dataset with output ports. Please Note: Subclasses of IterableDataset should *not* implement
    input_types.
    """

    @staticmethod
    def _collate_fn(batch):
        """A default implementation of a collation function. Users should override this method to define custom data
        loaders.
        """
        return data.dataloader.default_collate(batch)

    @typecheck()
    def collate_fn(self, batch):
        """This is the method that user pass as functor to DataLoader. The method optionally performs neural type
        checking and add types to the outputs.

        # Usage:

        .. code-block::

            dataloader = torch.utils.data.DataLoader(
                    ....,
                    collate_fn=dataset.collate_fn,
                    ....
            )

        Returns
        -------
        Collated batch, with or without types.
        """
        if self.input_types is not None:
            raise TypeError("Datasets should not implement `input_types` as they are not checked")

        # Simply forward the inner `_collate_fn`
        return self._collate_fn(batch)


@dataclass
class DatasetConfig:
    """Dataset configuration."""

    batch_size: int = 32
    drop_last: bool = False
    shuffle: bool = False
    num_workers: Optional[int] = 0
    pin_memory: bool = True
